#!/usr/bin/env python
# ***** BEGIN LICENSE BLOCK *****
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
# ***** END LICENSE BLOCK *****

import copy
import os
import sys
import signal
import socket
import subprocess
import telnetlib
import time
import tempfile

# load modules from parent dir
sys.path.insert(1, os.path.dirname(sys.path[0]))

from mozprocess import ProcessHandler

from mozharness.base.log import FATAL
from mozharness.base.script import BaseScript, PostScriptRun
from mozharness.base.vcs.vcsbase import VCSMixin
from mozharness.mozilla.blob_upload import BlobUploadMixin, blobupload_config_options
from mozharness.mozilla.mozbase import MozbaseMixin
from mozharness.mozilla.buildbot import TBPL_WORST_LEVEL_TUPLE
from mozharness.mozilla.testing.testbase import TestingMixin, testing_config_options
from mozharness.mozilla.testing.unittest import EmulatorMixin

from mozharness.mozilla.testing.device import ADBDeviceHandler


class AndroidEmulatorTest(BlobUploadMixin, TestingMixin, EmulatorMixin, VCSMixin, BaseScript, MozbaseMixin):
    config_options = [[
        ["--robocop-url"],
        {"action": "store",
         "dest": "robocop_url",
         "default": None,
         "help": "URL to the robocop apk",
         }
    ], [
        ["--host-utils-url"],
        {"action": "store",
         "dest": "xre_url",
         "default": None,
         "help": "URL to the host utils zip",
         }
    ], [
        ["--test-suite"],
        {"action": "append",
         "dest": "test_suites",
         }
    ], [
        ["--adb-path"],
        {"action": "store",
         "dest": "adb_path",
         "default": None,
         "help": "Path to adb",
         }
    ]] + copy.deepcopy(testing_config_options) + \
        copy.deepcopy(blobupload_config_options)

    error_list = [
    ]

    virtualenv_requirements = [
    ]

    virtualenv_modules = [
    ]

    app_name = None

    def __init__(self, require_config_file=False):
        super(AndroidEmulatorTest, self).__init__(
            config_options=self.config_options,
            all_actions=['clobber',
                         'read-buildbot-config',
                         'setup-avds',
                         'start-emulators',
                         'download-and-extract',
                         'create-virtualenv',
                         'install',
                         'run-tests',
                         'stop-emulators'],
            default_actions=['clobber',
                             'start-emulators',
                             'download-and-extract',
                             'create-virtualenv',
                             'install',
                             'run-tests',
                             'stop-emulators'],
            require_config_file=require_config_file,
            config={
                'virtualenv_modules': self.virtualenv_modules,
                'virtualenv_requirements': self.virtualenv_requirements,
                'require_test_zip': True,
                # IP address of the host as seen from the emulator
                'remote_webserver': '10.0.2.2',
            }
        )

        # these are necessary since self.config is read only
        c = self.config
        abs_dirs = self.query_abs_dirs()
        self.adb_path = self.query_exe('adb')
        self.installer_url = c.get('installer_url')
        self.installer_path = c.get('installer_path')
        self.test_url = c.get('test_url')
        self.test_manifest = c.get('test_manifest')
        self.robocop_url = c.get('robocop_url')
        self.robocop_path = os.path.join(abs_dirs['abs_work_dir'], "robocop.apk")
        self.host_utils_url = c.get('host_utils_url')
        self.minidump_stackwalk_path = c.get("minidump_stackwalk_path")
        self.emulators = c.get('emulators')
        self.test_suite_definitions = c['test_suite_definitions']
        self.test_suites = c.get('test_suites')
        for suite in self.test_suites:
            assert suite in self.test_suite_definitions

    def _query_tests_dir(self, suite_name):
        dirs = self.query_abs_dirs()
        suite_category = self.test_suite_definitions[suite_name]["category"]
        try:
            test_dir = self.config["suite_definitions"][suite_category]["testsdir"]
        except:
            test_dir = suite_category

        return os.path.join(dirs['abs_test_install_dir'], test_dir)

    def query_abs_dirs(self):
        if self.abs_dirs:
            return self.abs_dirs
        abs_dirs = super(AndroidEmulatorTest, self).query_abs_dirs()
        dirs = {}
        dirs['abs_test_install_dir'] = os.path.join(
            abs_dirs['abs_work_dir'], 'tests')
        dirs['abs_xre_dir'] = os.path.join(
            abs_dirs['abs_work_dir'], 'hostutils')
        dirs['abs_modules_dir'] = os.path.join(
            dirs['abs_test_install_dir'], 'modules')
        dirs['abs_blob_upload_dir'] = os.path.join(
            abs_dirs['abs_work_dir'], 'blobber_upload_dir')
        dirs['abs_emulator_dir'] = os.path.join(
            abs_dirs['abs_work_dir'], 'emulator')

        if self.config.get("developer_mode"):
            dirs['abs_avds_dir'] = os.path.join(
                abs_dirs['abs_work_dir'], "avds_dir")
        else:
            dirs['abs_avds_dir'] = "/home/cltbld/.android"

        for key in dirs.keys():
            if key not in abs_dirs:
                abs_dirs[key] = dirs[key]
        self.abs_dirs = abs_dirs
        return self.abs_dirs

    def _build_arg(self, option, value):
        """
        Build a command line argument
        """
        if not value:
            return []
        return [str(option), str(value)]

    def _redirectSUT(self, emulator_index):
        '''
        This redirects the default SUT ports for a given emulator.
        This is needed if more than one emulator is started.
        '''
        emulator = self.emulators[emulator_index]
        emuport = emulator["emulator_port"]
        sutport1 = emulator["sut_port1"]
        sutport2 = emulator["sut_port2"]
        attempts = 0
        tn = None
        redirect_completed = False
        while attempts < 5:
            if attempts == 0:
                self.info("Sleeping 10 seconds")
                time.sleep(10)
            else:
                self.info("Sleeping 30 seconds")
                time.sleep(30)
            attempts += 1
            self.info("  Attempt #%d to redirect ports: (%d, %d, %d)" %
                      (attempts, emuport, sutport1, sutport2))
            try:
                tn = telnetlib.Telnet('localhost', emuport, 300)
                break
            except socket.error, e:
                self.info("Trying again after exception: %s" % str(e))
                pass

        if tn is not None:
            res = tn.read_until('OK')
            if res.find('OK') == -1:
                self.warning('initial OK prompt not received from emulator: ' + str(res))
            tn.write('redir add tcp:' + str(sutport1) + ':' + str(self.config["default_sut_port1"]) + '\n')
            tn.write('redir add tcp:' + str(sutport2) + ':' + str(self.config["default_sut_port2"]) + '\n')
            tn.write('quit\n')
            res = tn.read_all()
            if res.find('OK') == -1:
                self.warning('error adding redirect: ' + str(res))
            else:
                redirect_completed = True
        else:
            self.warning('failed to establish a telnet connection with the emulator')
        return redirect_completed

    def _launch_emulator(self, emulator_index):
        emulator = self.emulators[emulator_index]
        env = self.query_env()

        # Set $LD_LIBRARY_PATH to self.dirs['abs_work_dir'] so that
        # the emulator picks up the symlink to libGL.so.1 that we
        # constructed in start_emulators.
        env['LD_LIBRARY_PATH'] = self.abs_dirs['abs_work_dir']

        # Specifically for developer mode we have to set this variable
        avd_home_dir = self.abs_dirs['abs_avds_dir']
        env['ANDROID_AVD_HOME'] = os.path.join(avd_home_dir, 'avd')

        command = [
            "emulator", "-avd", emulator["name"],
            "-port", str(emulator["emulator_port"]),
        ]
        if "emulator_extra_args" in self.config:
            command += self.config["emulator_extra_args"].split()

        tmp_file = tempfile.NamedTemporaryFile(mode='w')
        tmp_stdout = open(tmp_file.name, 'w')
        self.info("Created temp file %s." % tmp_file.name)
        self.info("Trying to start the emulator with this command: %s" % ' '.join(command))
        proc = subprocess.Popen(command, stdout=tmp_stdout, stderr=tmp_stdout, env=env)
        return {
            "process": proc,
            "tmp_file": tmp_file,
            "tmp_stdout": tmp_stdout
        }

    def _check_emulator(self, emulator):
        self.info('Checking emulator %s' % emulator["name"])

        if self.config["device_manager"] == "sut":
            attempts = 0
            tn = None
            contacted_sut = False
            while attempts < 8 and not contacted_sut:
                if attempts != 0:
                    self.info("Sleeping 30 seconds")
                    time.sleep(30)
                attempts += 1
                self.info("  Attempt #%d to connect to SUT on port %d" %
                          (attempts, emulator["sut_port1"]))
                try:
                    tn = telnetlib.Telnet('localhost', emulator["sut_port1"], 10)
                    if tn is not None:
                        self.info('Connected to port %d' % emulator["sut_port1"])
                        res = tn.read_until('$>', 10)
                        tn.write('quit\n')
                        if res.find('$>') == -1:
                            self.warning('Unexpected SUT response: %s' % res)
                        else:
                            self.info('SUT response: %s' % res)
                            contacted_sut = True
                        tn.read_all()
                    else:
                        self.warning('Unable to connect to the SUT agent on port %d' % emulator["sut_port1"])
                except socket.error, e:
                    self.info('Trying again after socket error: %s' % str(e))
                    pass
                except EOFError:
                    self.info('Trying again after EOF')
                    pass
                except:
                    self.info('Trying again after unexpected exception')
                    pass
                finally:
                    if tn is not None:
                        tn.close()
            if not contacted_sut:
                self.warning('Unable to communicate with SUT agent on port %d' % emulator["sut_port1"])

        attempts = 0
        tn = None
        contacted_emu = False
        while attempts < 4:
            if attempts != 0:
                self.info("Sleeping 30 seconds")
                time.sleep(30)
            attempts += 1
            self.info("  Attempt #%d to connect to emulator on port %d" %
                      (attempts, emulator["emulator_port"]))
            try:
                tn = telnetlib.Telnet('localhost', emulator["emulator_port"], 10)
                if tn is not None:
                    self.info('Connected to port %d' % emulator["emulator_port"])
                    res = tn.read_until('OK', 10)
                    self.info(res)
                    tn.write('avd status\n')
                    res = tn.read_until('OK', 10)
                    self.info('avd status: %s' % res)
                    tn.write('redir list\n')
                    res = tn.read_until('OK', 10)
                    self.info('redir list: %s' % res)
                    tn.write('network status\n')
                    res = tn.read_until('OK', 10)
                    self.info('network status: %s' % res)
                    tn.write('quit\n')
                    tn.read_all()
                    tn.close()
                    contacted_emu = True
                    break
                else:
                    self.warning('Unable to connect to the emulator on port %d' % emulator["emulator_port"])
            except socket.error, e:
                self.info('Trying again after socket error: %s' % str(e))
                pass
            except EOFError:
                self.info('Trying again after EOF')
                pass
            except:
                self.info('Trying again after unexpected exception')
                pass
            finally:
                if tn is not None:
                    tn.close()
        if not contacted_emu:
            self.warning('Unable to communicate with emulator on port %d' % emulator["emulator_port"])

        ps_cmd = [self.adb_path, '-s', emulator["device_id"], 'shell', 'ps']
        p = subprocess.Popen(ps_cmd, stdout=subprocess.PIPE)
        out, err = p.communicate()
        self.info('%s:\n%s\n%s' % (ps_cmd, out, err))

    def _dump_host_state(self):
        p = subprocess.Popen(['ps', '-ef'], stdout=subprocess.PIPE)
        out, err = p.communicate()
        self.info("output from ps -ef follows...")
        if out:
            self.info(out)
        if err:
            self.info(err)
        p = subprocess.Popen(['netstat', '-a', '-p', '-n', '-t', '-u'], stdout=subprocess.PIPE)
        out, err = p.communicate()
        self.info("output from netstat -a -p -n -t -u follows...")
        if out:
            self.info(out)
        if err:
            self.info(err)

    def _dump_emulator_log(self, emulator_index):
        emulator = self.emulators[emulator_index]
        self.info("##### %s emulator log begins" % emulator["name"])
        output = self.read_from_file(self.emulator_procs[emulator_index]["tmp_file"].name, verbose=False)
        if output:
            self.info(output)
        self.info("##### %s emulator log ends" % emulator["name"])

    def _kill_processes(self, process_name):
        p = subprocess.Popen(['ps', '-A'], stdout=subprocess.PIPE)
        out, err = p.communicate()
        self.info("Let's kill every process called %s" % process_name)
        for line in out.splitlines():
            if process_name in line:
                pid = int(line.split(None, 1)[0])
                self.info("Killing pid %d." % pid)
                os.kill(pid, signal.SIGKILL)

    @PostScriptRun
    def _post_script(self):
        self._kill_processes(self.config["emulator_process_name"])

    # XXX: This and android_panda.py's function might make sense to take higher up
    def _download_robocop_apk(self):
        dirs = self.query_abs_dirs()
        self.apk_url = self.installer_url[:self.installer_url.rfind('/')]
        robocop_url = self.apk_url + '/robocop.apk'
        self.info("Downloading robocop...")
        self.download_file(robocop_url, 'robocop.apk', dirs['abs_work_dir'], error_level=FATAL)

    def _query_package_name(self):
        if self.app_name is None:
            #find appname from package-name.txt - assumes download-and-extract has completed successfully
            apk_dir = self.abs_dirs['abs_work_dir']
            self.apk_path = os.path.join(apk_dir, self.installer_path)
            unzip = self.query_exe("unzip")
            package_path = os.path.join(apk_dir, 'package-name.txt')
            unzip_cmd = [unzip, '-q', '-o',  self.apk_path]
            self.run_command(unzip_cmd, cwd=apk_dir, halt_on_failure=True)
            self.app_name = str(self.read_from_file(package_path, verbose=True)).rstrip()
        return self.app_name

    def preflight_install(self):
        # in the base class, this checks for mozinstall, but we don't use it
        pass

    def _build_command(self, emulator, suite_name):
        c = self.config
        dirs = self.query_abs_dirs()
        suite_category = self.test_suite_definitions[suite_name]["category"]

        if suite_category not in c["suite_definitions"]:
            self.fatal("Key '%s' not defined in the config!" % suite_category)

        cmd = [
            self.query_python_path('python'),
            '-u',
            os.path.join(
                self._query_tests_dir(suite_name),
                c["suite_definitions"][suite_category]["run_filename"]
            ),
        ]

        raw_log_file = os.path.join(dirs['abs_blob_upload_dir'],
                                    '%s_raw.log' % suite_name)
        error_summary_file = os.path.join(dirs['abs_blob_upload_dir'],
                                          '%s_errorsummary.log' % suite_name)
        str_format_values = {
            'app': self._query_package_name(),
            'remote_webserver': c['remote_webserver'],
            'xre_path': os.path.join(dirs['abs_xre_dir'], 'xre'),
            'utility_path':  os.path.join(dirs['abs_xre_dir'], 'bin'),
            'http_port': emulator['http_port'],
            'ssl_port': emulator['ssl_port'],
            'certs_path': os.path.join(dirs['abs_work_dir'], 'tests/certs'),
            # TestingMixin._download_and_extract_symbols() will set
            # self.symbols_path when downloading/extracting.
            'symbols_path': self.symbols_path,
            'modules_dir': dirs['abs_modules_dir'],
            'installer_path': self.installer_path,
            'raw_log_file': raw_log_file,
            'error_summary_file': error_summary_file,
            'dm_trans': c['device_manager'],
        }
        if self.config["device_manager"] == "sut":
            str_format_values.update({
                'device_ip': c['device_ip'],
                'device_port': str(emulator['sut_port1']),
            })
        for option in c["suite_definitions"][suite_category]["options"]:
            cmd.extend([option % str_format_values])

        for arg in self.test_suite_definitions[suite_name]["extra_args"]:
            argname = arg.split('=')[0]
            # only add the extra arg if it wasn't already defined by in-tree configs
            if any(a.split('=')[0] == argname for a in cmd):
                continue
            cmd.append(arg)

        try_options, try_tests = self.try_args(suite_category)
        cmd.extend(try_options)
        cmd.extend(self.query_tests_args(
            self.config["suite_definitions"][suite_category].get("tests"),
            self.test_suite_definitions[suite_name].get("tests"),
            try_tests))

        return cmd

    def preflight_run_tests(self):
        super(AndroidEmulatorTest, self).preflight_run_tests()

        if not os.path.isfile(self.adb_path):
            self.fatal("The adb binary '%s' is not a valid file!" % self.adb_path)

    def _trigger_test(self, suite_name, emulator_index):
        """
        Run a test suite on an emulator

        We return a dictionary with the following information:
         - subprocess object that is running the test on the emulator
         - the filename where the stdout is going to
         - the stdout where the output is going to
         - the suite name that is associated
        """
        dirs = self.query_abs_dirs()
        cmd = self._build_command(self.emulators[emulator_index], suite_name)

        try:
            cwd = self._query_tests_dir(suite_name)
        except:
            self.fatal("Don't know how to run --test-suite '%s'!" % suite_name)

        env = self.query_env()
        if self.query_minidump_stackwalk():
            env['MINIDUMP_STACKWALK'] = self.minidump_stackwalk_path
        env['MOZ_UPLOAD_DIR'] = self.query_abs_dirs()['abs_blob_upload_dir']
        env['MINIDUMP_SAVE_PATH'] = self.query_abs_dirs()['abs_blob_upload_dir']

        self.info("Running on %s the command %s" % (self.emulators[emulator_index]["name"], subprocess.list2cmdline(cmd)))
        tmp_file = tempfile.NamedTemporaryFile(mode='w')
        tmp_stdout = open(tmp_file.name, 'w')
        self.info("Created temp file %s." % tmp_file.name)
        return {
            "process": subprocess.Popen(cmd, cwd=cwd, stdout=tmp_stdout, stderr=tmp_stdout, env=env),
            "tmp_file": tmp_file,
            "tmp_stdout": tmp_stdout,
            "suite_name": suite_name,
            "emulator_index": emulator_index
        }

    def _tooltool_fetch(self, url):
        c = self.config
        dirs = self.query_abs_dirs()

        manifest_path = self.download_file(
            url,
            file_name='releng.manifest',
            parent_dir=dirs['abs_avds_dir']
        )

        if not os.path.exists(manifest_path):
            self.fatal("Could not retrieve manifest needed to retrieve avds "
                       "artifacts from %s" % manifest_path)

        self.tooltool_fetch(manifest_path,
                            output_dir=dirs['abs_avds_dir'],
                            cache=c.get("tooltool_cache", None))

    ##########################################
    ### Actions for AndroidEmulatorTest ###
    ##########################################
    def setup_avds(self):
        '''
        If tooltool cache mechanism is enabled, the cached version is used by
        the fetch command. If the manifest includes an "unpack" field, tooltool
        will unpack all compressed archives mentioned in the manifest.
        '''
        c = self.config
        dirs = self.query_abs_dirs()

        # FIXME
        # Clobbering and re-unpacking would not be needed if we had a way to
        # check whether the unpacked content already present match the
        # contents of the tar ball
        self.rmtree(dirs['abs_avds_dir'])
        self.mkdir_p(dirs['abs_avds_dir'])
        if not self.buildbot_config:
            # XXX until we figure out how to determine the repo_path, revision
            url = 'https://hg.mozilla.org/%s/raw-file/%s/%s' % (
                "try", "default", c["tooltool_manifest_path"])
            self._tooltool_fetch(url)
        elif self.buildbot_config and 'properties' in self.buildbot_config:
            url = 'https://hg.mozilla.org/%s/raw-file/%s/%s' % (
                self.buildbot_config['properties']['repo_path'],
                self.buildbot_config['properties']['revision'],
                c["tooltool_manifest_path"])
            self._tooltool_fetch(url)
        else:
            self.fatal("properties in self.buildbot_config are required to "
                       "retrieve tooltool manifest to be used for avds setup")

        if self.config.get("developer_mode"):
            # For developer mode we have to modify the downloaded avds to
            # point to the right directory.
            avd_home_dir = self.abs_dirs['abs_avds_dir']
            cmd = [
                'bash', '-c',
                'sed -i "s|/home/cltbld/.android|%s|" %s/test-*.ini' %
                (avd_home_dir, os.path.join(avd_home_dir, 'avd'))
            ]
            proc = ProcessHandler(cmd)
            proc.run()
            proc.wait()


    def start_emulators(self):
        '''
        This action starts the emulators and redirects the two SUT ports for each one of them
        '''
        assert len(self.test_suites) <= len(self.emulators), \
            "We can't run more tests that the number of emulators we start"

        if 'emulator_url' in self.config or 'emulator_manifest' in self.config or 'tools_manifest' in self.config:
            self.install_emulator()

        if not self.config.get("developer_mode"):
            # We kill compiz because it sometimes prevents us from starting the emulators
            self._kill_processes("compiz")
            self._kill_processes("xpcshell")

        # We add a symlink for libGL.so because the emulator dlopen()s it by that name
        # even though the installed library on most systems without dev packages is
        # libGL.so.1
        linkfile = os.path.join(self.abs_dirs['abs_work_dir'], "libGL.so")
        self.info("Attempting to establish symlink for %s" % linkfile)
        try:
            os.unlink(linkfile)
        except OSError:
            pass
        for libdir in ["/usr/lib/x86_64-linux-gnu/mesa",
                       "/usr/lib/i386-linux-gnu/mesa",
                       "/usr/lib/mesa"]:
            libfile = os.path.join(libdir, "libGL.so.1")
            if os.path.exists(libfile):
                self.info("Symlinking %s -> %s" % (linkfile, libfile))
                self.mkdir_p(self.abs_dirs['abs_work_dir'])
                os.symlink(libfile, linkfile)
                break

        attempts = 0
        redirect_failed = True
        # Launch the required emulators and redirect the SUT ports for each. If unable
        # to redirect the SUT ports, kill the emulators and try starting them again.
        # The wait-and-retry logic is necessary because the emulators intermittently fail
        # to respond to telnet connections immediately after startup: bug 949740. In this
        # case, the emulator log shows "ioctl(KVM_CREATE_VM) failed: Interrupted system call".
        # We do not know how to avoid this error and the only way we have found to
        # recover is to kill the emulator and start again.
        while attempts < 3 and redirect_failed:
            if attempts > 0:
                self.info("Sleeping 30 seconds before retry")
                time.sleep(30)
            attempts += 1
            self.info('Attempt #%d to launch emulators...' % attempts)
            self._dump_host_state()
            self.emulator_procs = []
            emulator_index = 0
            redirect_failed = False
            for test in self.test_suites:
                emulator_proc = self._launch_emulator(emulator_index)
                self.emulator_procs.append(emulator_proc)
                if self.config["device_manager"] == "sut":
                    if self._redirectSUT(emulator_index):
                        emulator = self.emulators[emulator_index]
                        self.info("%s: %s; sut port: %s/%s" %
                                  (emulator["name"], emulator["emulator_port"], emulator["sut_port1"], emulator["sut_port2"]))
                        emulator_index += 1
                    else:
                        self._dump_emulator_log(emulator_index)
                        self._kill_processes(self.config["emulator_process_name"])
                        redirect_failed = True
                        break
        if redirect_failed:
            self.fatal('We have not been able to establish a telnet connection with the emulator')

        # Verify that we can communicate with each emulator
        emulator_index = 0
        for test in self.test_suites:
            emulator = self.emulators[emulator_index]
            emulator_index += 1
            self._check_emulator(emulator)
        # Start logcat for each emulator. Each adb process runs until the
        # corresponding emulator is killed. Output is written directly to
        # the blobber upload directory so that it is uploaded automatically
        # at the end of the job.
        self.mkdir_p(self.abs_dirs['abs_blob_upload_dir'])
        emulator_index = 0
        for test in self.test_suites:
            emulator = self.emulators[emulator_index]
            emulator_index += 1
            logcat_filename = 'logcat-%s.log' % emulator["device_id"]
            logcat_path = os.path.join(self.abs_dirs['abs_blob_upload_dir'], logcat_filename)
            logcat_cmd = '%s -s %s logcat -v time Trace:S StrictMode:S ExchangeService:S > %s &' % \
                (self.adb_path, emulator["device_id"], logcat_path)
            self.info(logcat_cmd)
            os.system(logcat_cmd)
        # Create the /data/anr directory on each emulator image.
        emulator_index = 0
        for test in self.test_suites:
            emulator = self.emulators[emulator_index]
            emulator_index += 1
            mkdir_cmd = [self.adb_path, '-s', emulator["device_id"], 'shell', 'mkdir', '/data/anr']
            p = subprocess.Popen(mkdir_cmd, stdout=subprocess.PIPE)
            out, err = p.communicate()
            self.info('%s:\n%s\n%s' % (mkdir_cmd, out, err))

    def download_and_extract(self):
        # This will download and extract the fennec.apk and tests.zip
        suite_categories = set([self.test_suite_definitions[c]['category']
                                for c in self.test_suites])
        super(AndroidEmulatorTest, self).download_and_extract(suite_categories=suite_categories)
        dirs = self.query_abs_dirs()

        for suite_name in self.test_suites:
            if suite_name.startswith('robocop'):
                self._download_robocop_apk()
                break

        self.download_unzip(self.host_utils_url, dirs['abs_xre_dir'])

    def install(self):
        assert self.installer_path is not None, \
            "Either add installer_path to the config or use --installer-path."

        emulator_index = 0
        for suite_name in self.test_suites:
            emulator = self.emulators[emulator_index]
            emulator_index += 1

            config = {
                'device-id': emulator["device_id"],
                'enable_automation': True,
                'device_package_name': self._query_package_name()
            }
            config = dict(config.items() + self.config.items())

            self.info("Creating ADBDevicHandler for %s with config %s" % (emulator["name"], config))
            dh = ADBDeviceHandler(config=config, log_obj=self.log_obj, script_obj=self)
            dh.device_id = emulator['device_id']

            # Wait for Android to finish booting
            completed = None
            retries = 0
            while retries < 30:
                completed = self.get_output_from_command([self.adb_path,
                    "-s", emulator['device_id'], "shell",
                    "getprop", "sys.boot_completed"])
                if completed == '1':
                    break
                time.sleep(10)
                retries = retries + 1
            if completed != '1':
                self.warning('Retries exhausted waiting for Android boot.')

            # Install Fennec
            self.info("Installing Fennec for %s" % emulator["name"])
            dh.install_app(self.installer_path)

            # Install the robocop apk if required
            if suite_name.startswith('robocop'):
                self.info("Installing Robocop for %s" % emulator["name"])
                config['device_package_name'] = self.config["robocop_package_name"]
                dh.install_app(self.robocop_path)

            self.info("Finished installing apps for %s" % emulator["name"])

    def run_tests(self):
        """
        Run the tests
        """
        procs = []

        emulator_index = 0
        for suite_name in self.test_suites:
            procs.append(self._trigger_test(suite_name, emulator_index))
            emulator_index += 1

        joint_tbpl_status = None
        joint_log_level = None
        start_time = int(time.time())
        while True:
            for p in procs:
                emulator_index = p["emulator_index"]
                return_code = p["process"].poll()
                if return_code is not None:
                    suite_name = p["suite_name"]
                    # To make reading the log of the suite not mix with the previous line
                    sys.stdout.write('\n')
                    self.info("##### %s log begins" % p["suite_name"])
                    # Let's close the stdout
                    p["tmp_stdout"].close()
                    # Let's read the file that now has the output
                    output = self.read_from_file(p["tmp_file"].name, verbose=False)
                    # Let's parse the output (which also prints it)
                    # and determine what the results should be
                    parser = self.get_test_output_parser(
                        self.test_suite_definitions[p["suite_name"]]["category"],
                        config=self.config,
                        log_obj=self.log_obj,
                        error_list=self.error_list)
                    for line in output.splitlines():
                        parser.parse_single_line(line)

                    # After parsing each line we should know what the summary for this suite should be
                    tbpl_status, log_level = parser.evaluate_parser(0)
                    parser.append_tinderboxprint_line(p["suite_name"])
                    # After running all jobs we will report the worst status of all emulator runs
                    joint_tbpl_status = self.worst_level(tbpl_status, joint_tbpl_status, TBPL_WORST_LEVEL_TUPLE)
                    joint_log_level = self.worst_level(log_level, joint_log_level)

                    self.info("##### %s log ends" % p["suite_name"])
                    self._dump_emulator_log(emulator_index)
                    procs.remove(p)
            if procs == []:
                break
            else:
                # Every 5 minutes let's print something to stdout
                # so buildbot won't kill the process due to lack of output
                if int(time.time()) - start_time > 5 * 60:
                    self.info('#')
                    start_time = int(time.time())
                time.sleep(30)

        self.buildbot_status(joint_tbpl_status, level=joint_log_level)

    def stop_emulators(self):
        '''
        Report emulator health, then make sure that every emulator has been stopped
        '''
        emulator_index = 0
        for test in self.test_suites:
            emulator = self.emulators[emulator_index]
            emulator_index += 1
            self._check_emulator(emulator)
        self._kill_processes(self.config["emulator_process_name"])

if __name__ == '__main__':
    emulatorTest = AndroidEmulatorTest()
    emulatorTest.run_and_exit()



#!/usr/bin/env python
# ***** BEGIN LICENSE BLOCK *****
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
# ***** END LICENSE BLOCK *****
'''Python usage, esp. virtualenv.
'''

import os
import subprocess
import sys
import time
import json
import traceback

from mozharness.base.script import (
    PostScriptAction,
    PostScriptRun,
    PreScriptAction,
    PreScriptRun,
)
from mozharness.base.errors import VirtualenvErrorList
from mozharness.base.log import WARNING, FATAL
from mozharness.mozilla.proxxy import Proxxy

def get_tlsv1_post():
    # Monkeypatch to work around SSL errors in non-bleeding-edge Python.
    # Taken from https://lukasa.co.uk/2013/01/Choosing_SSL_Version_In_Requests/
    import requests
    from requests.packages.urllib3.poolmanager import PoolManager
    import ssl

    class TLSV1Adapter(requests.adapters.HTTPAdapter):
        def init_poolmanager(self, connections, maxsize, block=False):
            self.poolmanager = PoolManager(num_pools=connections,
                                           maxsize=maxsize,
                                           block=block,
                                           ssl_version=ssl.PROTOCOL_TLSv1)
    s = requests.Session()
    s.mount('https://', TLSV1Adapter())
    return s.post

# Virtualenv {{{1
virtualenv_config_options = [
    [["--venv-path", "--virtualenv-path"], {
        "action": "store",
        "dest": "virtualenv_path",
        "default": "venv",
        "help": "Specify the path to the virtualenv top level directory"
    }],
    [["--virtualenv"], {
        "action": "store",
        "dest": "virtualenv",
        "help": "Specify the virtualenv executable to use"
    }],
    [["--find-links"], {
        "action": "extend",
        "dest": "find_links",
        "help": "URL to look for packages at"
    }],
    [["--pip-index"], {
        "action": "store_true",
        "default": True,
        "dest": "pip_index",
        "help": "Use pip indexes (default)"
    }],
    [["--no-pip-index"], {
        "action": "store_false",
        "dest": "pip_index",
        "help": "Don't use pip indexes"
    }],
]


class VirtualenvMixin(object):
    '''BaseScript mixin, designed to create and use virtualenvs.

    Config items:
     * virtualenv_path points to the virtualenv location on disk.
     * virtualenv_modules lists the module names.
     * MODULE_url list points to the module URLs (optional)
    Requires virtualenv to be in PATH.
    Depends on ScriptMixin
    '''
    python_paths = {}
    site_packages_path = None

    def __init__(self, *args, **kwargs):
        self._virtualenv_modules = []
        super(VirtualenvMixin, self).__init__(*args, **kwargs)

    def register_virtualenv_module(self, name=None, url=None, method=None,
                                   requirements=None, optional=False,
                                   two_pass=False, editable=False):
        """Register a module to be installed with the virtualenv.

        This method can be called up until create_virtualenv() to register
        modules that should be installed in the virtualenv.

        See the documentation for install_module for how the arguments are
        applied.
        """
        self._virtualenv_modules.append((name, url, method, requirements,
                                         optional, two_pass, editable))

    def query_virtualenv_path(self):
        c = self.config
        dirs = self.query_abs_dirs()
        virtualenv = None
        if 'abs_virtualenv_dir' in dirs:
            virtualenv = dirs['abs_virtualenv_dir']
        elif c.get('virtualenv_path'):
            if os.path.isabs(c['virtualenv_path']):
                virtualenv = c['virtualenv_path']
            else:
                virtualenv = os.path.join(dirs['abs_work_dir'],
                                          c['virtualenv_path'])
        return virtualenv

    def query_python_path(self, binary="python"):
        """Return the path of a binary inside the virtualenv, if
        c['virtualenv_path'] is set; otherwise return the binary name.
        Otherwise return None
        """
        if binary not in self.python_paths:
            bin_dir = 'bin'
            if self._is_windows():
                bin_dir = 'Scripts'
            virtualenv_path = self.query_virtualenv_path()
            if virtualenv_path:
                self.python_paths[binary] = os.path.abspath(os.path.join(virtualenv_path, bin_dir, binary))
            else:
                self.python_paths[binary] = self.query_exe(binary)
        return self.python_paths[binary]

    def query_python_site_packages_path(self):
        if self.site_packages_path:
            return self.site_packages_path
        python = self.query_python_path()
        self.site_packages_path = self.get_output_from_command(
            [python, '-c',
             'from distutils.sysconfig import get_python_lib; ' +
             'print(get_python_lib())'])
        return self.site_packages_path

    def package_versions(self, pip_freeze_output=None, error_level=WARNING, log_output=False):
        """
        reads packages from `pip freeze` output and returns a dict of
        {package_name: 'version'}
        """
        packages = {}

        if pip_freeze_output is None:
            # get the output from `pip freeze`
            pip = self.query_python_path("pip")
            if not pip:
                self.log("package_versions: Program pip not in path", level=error_level)
                return {}
            pip_freeze_output = self.get_output_from_command([pip, "freeze"], silent=True)
            if not isinstance(pip_freeze_output, basestring):
                self.fatal("package_versions: Error encountered running `pip freeze`: %s" % pip_freeze_output)

        for line in pip_freeze_output.splitlines():
            # parse the output into package, version
            line = line.strip()
            if not line:
                # whitespace
                continue
            if line.startswith('-'):
                # not a package, probably like '-e http://example.com/path#egg=package-dev'
                continue
            if '==' not in line:
                self.fatal("pip_freeze_packages: Unrecognized output line: %s" % line)
            package, version = line.split('==', 1)
            packages[package] = version

        if log_output:
            self.info("Current package versions:")
            for package in sorted(packages):
                self.info("  %s == %s" % (package, packages[package]))

        return packages

    def is_python_package_installed(self, package_name, error_level=WARNING):
        """
        Return whether the package is installed
        """
        packages = self.package_versions(error_level=error_level).keys()
        return package_name.lower() in [package.lower() for package in packages]

    def install_module(self, module=None, module_url=None, install_method=None,
                       requirements=(), optional=False, global_options=[],
                       no_deps=False, editable=False):
        """
        Install module via pip.

        module_url can be a url to a python package tarball, a path to
        a directory containing a setup.py (absolute or relative to work_dir)
        or None, in which case it will default to the module name.

        requirements is a list of pip requirements files.  If specified, these
        will be combined with the module_url (if any), like so:

        pip install -r requirements1.txt -r requirements2.txt module_url
        """
        c = self.config
        dirs = self.query_abs_dirs()
        env = self.query_env()
        venv_path = self.query_virtualenv_path()
        self.info("Installing %s into virtualenv %s" % (module, venv_path))
        if not module_url:
            module_url = module
        if install_method in (None, 'pip'):
            if not module_url and not requirements:
                self.fatal("Must specify module and/or requirements")
            pip = self.query_python_path("pip")
            if c.get("verbose_pip"):
                command = [pip, "-v", "install"]
            else:
                command = [pip, "install"]
            if no_deps:
                command += ["--no-deps"]
            # To avoid timeouts with our pypi server, increase default timeout:
            # https://bugzilla.mozilla.org/show_bug.cgi?id=1007230#c802
            command += ['--timeout', str(c.get('pip_timeout', 120))]
            for requirement in requirements:
                command += ["-r", requirement]
            if c.get('find_links') and not c["pip_index"]:
                command += ['--no-index']
            for opt in global_options:
                command += ["--global-option", opt]
        elif install_method == 'easy_install':
            if not module:
                self.fatal("module parameter required with install_method='easy_install'")
            if requirements:
                # Install pip requirements files separately, since they're
                # not understood by easy_install.
                self.install_module(requirements=requirements,
                                    install_method='pip')
            # Allow easy_install to be overridden by
            # self.config['exes']['easy_install']
            default = 'easy_install'
            if self._is_windows():
                # Don't invoke `easy_install` directly on windows since
                # the 'install' in the executable name hits UAC
                # - http://answers.microsoft.com/en-us/windows/forum/windows_7-security/uac-message-do-you-want-to-allow-the-following/bea30ad8-9ef8-4897-aab4-841a65f7af71
                # - https://bugzilla.mozilla.org/show_bug.cgi?id=791840
                default = [self.query_python_path(), self.query_python_path('easy_install-script.py')]
            command = self.query_exe('easy_install', default=default, return_type="list")
        else:
            self.fatal("install_module() doesn't understand an install_method of %s!" % install_method)

        # Add --find-links pages to look at
        proxxy = Proxxy(self.config, self.log_obj)
        for link in proxxy.get_proxies_and_urls(c.get('find_links', [])):
            command.extend(["--find-links", link])

        # module_url can be None if only specifying requirements files
        if module_url:
            if editable:
                if install_method in (None, 'pip'):
                    command += ['-e']
                else:
                    self.fatal("editable installs not supported for install_method %s" % install_method)
            command += [module_url]

        # If we're only installing a single requirements file, use
        # the file's directory as cwd, so relative paths work correctly.
        cwd = dirs['abs_work_dir']
        if not module and len(requirements) == 1:
            cwd = os.path.dirname(requirements[0])

        quoted_command = subprocess.list2cmdline(command)
        # Allow for errors while building modules, but require a
        # return status of 0.
        self.retry(
            self.run_command,
            # None will cause default value to be used
            attempts=1 if optional else None,
            good_statuses=(0,),
            error_level=WARNING if optional else FATAL,
            error_message='Could not install python package: ' + quoted_command + ' failed after %(attempts)d tries!',
            args=[command, ],
            kwargs={
                'error_list': VirtualenvErrorList,
                'cwd': cwd,
                'env': env,
                # WARNING only since retry will raise final FATAL if all
                # retry attempts are unsuccessful - and we only want
                # an ERROR of FATAL if *no* retry attempt works
                'error_level': WARNING,
            }
        )

    def create_virtualenv(self, modules=(), requirements=()):
        """
        Create a python virtualenv.

        The virtualenv exe can be defined in c['virtualenv'] or
        c['exes']['virtualenv'], as a string (path) or list (path +
        arguments).

        c['virtualenv_python_dll'] is an optional config item that works
        around an old windows virtualenv bug.

        virtualenv_modules can be a list of module names to install, e.g.

            virtualenv_modules = ['module1', 'module2']

        or it can be a heterogeneous list of modules names and dicts that
        define a module by its name, url-or-path, and a list of its global
        options.

            virtualenv_modules = [
                {
                    'name': 'module1',
                    'url': None,
                    'global_options': ['--opt', '--without-gcc']
                },
                {
                    'name': 'module2',
                    'url': 'http://url/to/package',
                    'global_options': ['--use-clang']
                },
                {
                    'name': 'module3',
                    'url': os.path.join('path', 'to', 'setup_py', 'dir')
                    'global_options': []
                },
                'module4'
            ]

        virtualenv_requirements is an optional list of pip requirements files to
        use when invoking pip, e.g.,

            virtualenv_requirements = [
                '/path/to/requirements1.txt',
                '/path/to/requirements2.txt'
            ]
        """
        c = self.config
        dirs = self.query_abs_dirs()
        venv_path = self.query_virtualenv_path()
        self.info("Creating virtualenv %s" % venv_path)
        virtualenv = c.get('virtualenv', self.query_exe('virtualenv'))
        if isinstance(virtualenv, str):
            # allow for [python, virtualenv] in config
            virtualenv = [virtualenv]

        if not os.path.exists(virtualenv[0]) and not self.which(virtualenv[0]):
            self.add_summary("The executable '%s' is not found; not creating "
                             "virtualenv!" % virtualenv[0], level=FATAL)
            return -1

        # https://bugs.launchpad.net/virtualenv/+bug/352844/comments/3
        # https://bugzilla.mozilla.org/show_bug.cgi?id=700415#c50
        if c.get('virtualenv_python_dll'):
            # We may someday want to copy a differently-named dll, but
            # let's not think about that right now =\
            dll_name = os.path.basename(c['virtualenv_python_dll'])
            target = self.query_python_path(dll_name)
            scripts_dir = os.path.dirname(target)
            self.mkdir_p(scripts_dir)
            self.copyfile(c['virtualenv_python_dll'], target, error_level=WARNING)
        else:
            self.mkdir_p(dirs['abs_work_dir'])

        # make this list configurable?
        for module in ('distribute', 'pip'):
            if c.get('%s_url' % module):
                self.download_file(c['%s_url' % module],
                                   parent_dir=dirs['abs_work_dir'])

        virtualenv_options = c.get('virtualenv_options',
                                   ['--no-site-packages', '--distribute'])

        if os.path.exists(self.query_python_path()):
            self.info("Virtualenv %s appears to already exist; skipping virtualenv creation." % self.query_python_path())
        else:
            self.run_command(virtualenv + virtualenv_options + [venv_path],
                             cwd=dirs['abs_work_dir'],
                             error_list=VirtualenvErrorList,
                             halt_on_failure=True)
        if not modules:
            modules = c.get('virtualenv_modules', [])
        if not requirements:
            requirements = c.get('virtualenv_requirements', [])
        if not modules and requirements:
            self.install_module(requirements=requirements,
                                install_method='pip')
        for module in modules:
            module_url = module
            global_options = []
            if isinstance(module, dict):
                if module.get('name', None):
                    module_name = module['name']
                else:
                    self.fatal("Can't install module without module name: %s" %
                               str(module))
                module_url = module.get('url', None)
                global_options = module.get('global_options', [])
            else:
                module_url = self.config.get('%s_url' % module, module_url)
                module_name = module
            install_method = 'pip'
            if module_name in ('pywin32',):
                install_method = 'easy_install'
            self.install_module(module=module_name,
                                module_url=module_url,
                                install_method=install_method,
                                requirements=requirements,
                                global_options=global_options)

        for module, url, method, requirements, optional, two_pass, editable in \
                self._virtualenv_modules:
            if two_pass:
                self.install_module(
                    module=module, module_url=url,
                    install_method=method, requirements=requirements or (),
                    optional=optional, no_deps=True, editable=editable
                )
            self.install_module(
                module=module, module_url=url,
                install_method=method, requirements=requirements or (),
                optional=optional, editable=editable
            )

        self.info("Done creating virtualenv %s." % venv_path)

        self.package_versions(log_output=True)

    def activate_virtualenv(self):
        """Import the virtualenv's packages into this Python interpreter."""
        bin_dir = os.path.dirname(self.query_python_path())
        activate = os.path.join(bin_dir, 'activate_this.py')
        execfile(activate, dict(__file__=activate))


class ResourceMonitoringMixin(object):
    """Provides resource monitoring capabilities to scripts.

    When this class is in the inheritance chain, resource usage stats of the
    executing script will be recorded.

    This class requires the VirtualenvMixin in order to install a package used
    for recording resource usage.

    While we would like to record resource usage for the entirety of a script,
    since we require an external package, we can only record resource usage
    after that package is installed (as part of creating the virtualenv).
    That's just the way things have to be.
    """
    def __init__(self, *args, **kwargs):
        super(ResourceMonitoringMixin, self).__init__(*args, **kwargs)

        self.register_virtualenv_module('psutil>=0.7.1', method='pip',
                                        optional=True)
        self.register_virtualenv_module('mozsystemmonitor==0.0.0',
                                        method='pip', optional=True)
        self._resource_monitor = None

    @PostScriptAction('create-virtualenv')
    def _start_resource_monitoring(self, action, success=None):
        self.activate_virtualenv()

        # Resource Monitor requires Python 2.7, however it's currently optional.
        # Remove when all machines have had their Python version updated (bug 711299).
        if sys.version_info[:2] < (2, 7):
            self.warning('Resource monitoring will not be enabled! Python 2.7+ required.')
            return

        try:
            from mozsystemmonitor.resourcemonitor import SystemResourceMonitor

            self.info("Starting resource monitoring.")
            self._resource_monitor = SystemResourceMonitor(poll_interval=1.0)
            self._resource_monitor.start()
        except Exception:
            self.warning("Unable to start resource monitor: %s" %
                         traceback.format_exc())

    @PreScriptAction
    def _resource_record_pre_action(self, action):
        # Resource monitor isn't available until after create-virtualenv.
        if not self._resource_monitor:
            return

        self._resource_monitor.begin_phase(action)

    @PostScriptAction
    def _resource_record_post_action(self, action, success=None):
        # Resource monitor isn't available until after create-virtualenv.
        if not self._resource_monitor:
            return

        self._resource_monitor.finish_phase(action)

    @PostScriptRun
    def _resource_record_post_run(self):
        if not self._resource_monitor:
            return

        # This should never raise an exception. This is a workaround until
        # mozsystemmonitor is fixed. See bug 895388.
        try:
            self._resource_monitor.stop()
            self._log_resource_usage()
        except Exception:
            self.warning("Exception when reporting resource usage: %s" %
                         traceback.format_exc())

    def _log_resource_usage(self):
        rm = self._resource_monitor

        if rm.start_time is None:
            return

        def resources(phase):
            cpu_percent = rm.aggregate_cpu_percent(phase=phase, per_cpu=False)
            cpu_times = rm.aggregate_cpu_times(phase=phase, per_cpu=False)
            io = rm.aggregate_io(phase=phase)

            return cpu_percent, cpu_times, io

        def log_usage(prefix, duration, cpu_percent, cpu_times, io):
            message = '{prefix} - Wall time: {duration:.0f}s; ' \
                'CPU: {cpu_percent}; ' \
                'Read bytes: {io_read_bytes}; Write bytes: {io_write_bytes}; ' \
                'Read time: {io_read_time}; Write time: {io_write_time}'

            # XXX Some test harnesses are complaining about a string being
            # being fed into a 'f' formatter. This will help diagnose the
            # issue.
            cpu_percent_str = str(round(cpu_percent)) + '%' if cpu_percent else "Can't collect data"

            try:
                self.info(
                    message.format(
                        prefix=prefix, duration=duration,
                        cpu_percent=cpu_percent_str, io_read_bytes=io.read_bytes,
                        io_write_bytes=io.write_bytes, io_read_time=io.read_time,
                        io_write_time=io.write_time
                    )
                )
            except ValueError:
                self.warning("Exception when formatting: %s" %
                             traceback.format_exc())

        cpu_percent, cpu_times, io = resources(None)
        duration = rm.end_time - rm.start_time

        log_usage('Total resource usage', duration, cpu_percent, cpu_times, io)

        for phase in rm.phases.keys():
            start_time, end_time = rm.phases[phase]
            cpu_percent, cpu_times, io = resources(phase)
            log_usage(phase, end_time - start_time, cpu_percent, cpu_times, io)


class InfluxRecordingMixin(object):
    """Provides InfluxDB stat recording to scripts.

    This class records stats to an InfluxDB server, if enabled. Stat recording
    is enabled in a script by inheriting from this class, and adding an
    influxdb_credentials line to the influx_credentials_file (usually oauth.txt
    in automation).  This line should look something like:

        influxdb_credentials = 'http://goldiewilson-onepointtwentyone-1.c.influxdb.com:8086/db/DBNAME/series?u=DBUSERNAME&p=DBPASSWORD'

    Where DBNAME, DBUSERNAME, and DBPASSWORD correspond to the database name,
    and user/pw credentials for recording to the database. The stats from
    mozharness are recorded in the 'mozharness' table.
    """

    @PreScriptRun
    def influxdb_recording_init(self):
        self.recording = False
        self.post = None
        self.posturl = None
        self.build_metrics_summary = None
        self.res_props = self.config.get('build_resources_path') % self.query_abs_dirs()
        self.info("build_resources.json path: %s" % self.res_props)
        if self.res_props:
            self.rmtree(self.res_props)

        try:
            site_packages_path = self.query_python_site_packages_path()
            if site_packages_path not in sys.path:
                sys.path.append(site_packages_path)

            self.post = get_tlsv1_post()

            auth = os.path.join(os.getcwd(), self.config['influx_credentials_file'])
            if not os.path.exists(auth):
                self.warning("Unable to start influxdb recording: %s not found" % (auth,))
                return
            credentials = {}
            execfile(auth, credentials)
            if 'influxdb_credentials' in credentials:
                self.posturl = credentials['influxdb_credentials']
                self.recording = True
            else:
                self.warning("Unable to start influxdb recording: no credentials")
                return

        except Exception:
            # The exact reason for failing to start stats doesn't really matter.
            # If anything fails, we just won't record stats for this job.
            self.warning("Unable to start influxdb recording: %s" %
                         traceback.format_exc())
            return

    @PreScriptAction
    def influxdb_recording_pre_action(self, action):
        if not self.recording:
            return

        self.start_time = time.time()

    @PostScriptAction
    def influxdb_recording_post_action(self, action, success=None):
        if not self.recording:
            return

        elapsed_time = time.time() - self.start_time

        c = {}
        p = {}
        if self.buildbot_config:
            c = self.buildbot_config.get('properties', {})
        if self.buildbot_properties:
            p = self.buildbot_properties
        self.record_influx_stat([{
            "points": [[
                action,
                elapsed_time,
                c.get('buildername'),
                c.get('product'),
                c.get('platform'),
                c.get('branch'),
                c.get('slavename'),
                c.get('revision'),
                p.get('gaia_revision'),
                c.get('buildid'),
            ]],
            "name": "mozharness",
            "columns": [
                "action",
                "runtime",
                "buildername",
                "product",
                "platform",
                "branch",
                "slavename",
                "gecko_revision",
                "gaia_revision",
                "buildid",
            ],
        }])

    def _get_resource_usage(self, res, name, iolen, cpulen):
        c = {}
        p = {}
        if self.buildbot_config:
            c = self.buildbot_config.get('properties', {})
        if self.buildbot_properties:
            p = self.buildbot_properties

        data = [
            # Build properties
            c.get('buildername'),
            c.get('product'),
            c.get('platform'),
            c.get('branch'),
            c.get('slavename'),
            c.get('revision'),
            p.get('gaia_revision'),
            c.get('buildid'),

            # Mach step properties
            name,
            res.get('start'),
            res.get('end'),
            res.get('duration'),
            res.get('cpu_percent'),
        ]
        # The io and cpu_times fields are arrays, though they aren't always
        # present if a step completes before resource utilization is measured.
        # We add the arrays if they exist, otherwise we just do an array of None
        # to fill up the stat point.
        data.extend(res.get('io', [None] * iolen))
        data.extend(res.get('cpu_times', [None] * cpulen))
        return data

    @PostScriptAction('build')
    def record_mach_stats(self, action, success=None):
        if not os.path.exists(self.res_props):
            self.info('No build_resources.json found, not logging stats')
            return
        with open(self.res_props) as fh:
            resources = json.load(fh)
            data = {
                "points": [
                ],
                "name": "mach",
                "columns": [
                    # Build properties
                    "buildername",
                    "product",
                    "platform",
                    "branch",
                    "slavename",
                    "gecko_revision",
                    "gaia_revision",
                    "buildid",

                    # Mach step properties
                    "name",
                    "start",
                    "end",
                    "duration",
                    "cpu_percent",
                ],
            }

            # The io and cpu_times fields aren't static - they may vary based
            # on the specific platform being measured. Mach records the field
            # names, which we use as the column names here.
            data['columns'].extend(resources['io_fields'])
            data['columns'].extend(resources['cpu_times_fields'])
            iolen = len(resources['io_fields'])
            cpulen = len(resources['cpu_times_fields'])

            if 'duration' in resources:
                self.build_metrics_summary = {
                    'name': 'build times',
                    'value': resources['duration'],
                    'subtests': [],
                }

            # The top-level data has the overall resource usage, which we record
            # under the name 'TOTAL' to separate it from the individual tiers.
            data['points'].append(self._get_resource_usage(resources, 'TOTAL', iolen, cpulen))

            # Each tier also has the same resource stats as the top-level.
            for tier in resources['tiers']:
                data['points'].append(self._get_resource_usage(tier, tier['name'], iolen, cpulen))
                if 'duration' not in tier:
                    self.build_metrics_summary = None
                elif self.build_metrics_summary:
                    self.build_metrics_summary['subtests'].append({
                        'name': tier['name'],
                        'value': tier['duration'],
                    })

            self.record_influx_stat([data])

    def record_influx_stat(self, json_data):
        if not self.recording:
            return
        try:
            r = self.post(self.posturl, data=json.dumps(json_data), timeout=5)
            if r.status_code != 200:
                self.warning("Failed to log stats. Return code = %i, stats = %s" % (r.status_code, json_data))

                # Disable recording for the rest of this job. Even if it's just
                # intermittent, we don't want to keep the build from progressing.
                self.recording = False
        except Exception, e:
            self.warning('Failed to log stats. Exception = %s' % str(e))
            self.recording = False


# __main__ {{{1

if __name__ == '__main__':
    '''TODO: unit tests.
    '''
    pass



#!/usr/bin/env python
# ***** BEGIN LICENSE BLOCK *****
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
# ***** END LICENSE BLOCK *****
"""fx_desktop_build.py.

script harness to build nightly firefox within Mozilla's build environment
and developer machines alike

author: Jordan Lund

"""

import sys
import os

# load modules from parent dir
sys.path.insert(1, os.path.dirname(sys.path[0]))

from mozharness.mozilla.building.buildbase import BUILD_BASE_CONFIG_OPTIONS, \
    BuildingConfig, BuildScript


class FxDesktopBuild(BuildScript, object):
    def __init__(self):
        buildscript_kwargs = {
            'config_options': BUILD_BASE_CONFIG_OPTIONS,
            'all_actions': [
                'clobber',
                'clone-tools',
                'checkout-sources',
                'setup-mock',
                'build',
                'upload-files',  # upload from BB to TC
                'sendchange',
                'check-test',
                'package-source',
                'generate-source-signing-manifest',
                'multi-l10n',
                'generate-build-stats',
                'update',
            ],
            'require_config_file': True,
            # Default configuration
            'config': {
                'is_automation': True,
                "pgo_build": False,
                "debug_build": False,
                "pgo_platforms": ['linux', 'linux64', 'win32', 'win64'],
                # nightly stuff
                "nightly_build": False,
                'balrog_credentials_file': 'oauth.txt',
                'taskcluster_credentials_file': 'oauth.txt',
                'periodic_clobber': 168,
                # hg tool stuff
                'default_vcs': 'hgtool',
                "tools_repo": "https://hg.mozilla.org/build/tools",
                # Seed all clones with mozilla-unified. This ensures subsequent
                # jobs have a minimal `hg pull`.
                "clone_upstream_url": "https://hg.mozilla.org/mozilla-unified",
                "repo_base": "https://hg.mozilla.org",
                'tooltool_url': 'https://api.pub.build.mozilla.org/tooltool/',
                "graph_selector": "/server/collect.cgi",
                # only used for make uploadsymbols
                'old_packages': [
                    "%(objdir)s/dist/firefox-*",
                    "%(objdir)s/dist/fennec*",
                    "%(objdir)s/dist/seamonkey*",
                    "%(objdir)s/dist/thunderbird*",
                    "%(objdir)s/dist/install/sea/*.exe"
                ],
                'stage_product': 'firefox',
                'platform_supports_post_upload_to_latest': True,
                'use_branch_in_symbols_extra_buildid': True,
                'latest_mar_dir': '/pub/mozilla.org/firefox/nightly/latest-%(branch)s',
                'compare_locales_repo': 'https://hg.mozilla.org/build/compare-locales',
                'compare_locales_rev': 'RELEASE_AUTOMATION',
                'compare_locales_vcs': 'hgtool',
                'influx_credentials_file': 'oauth.txt',
                'build_resources_path': '%(abs_src_dir)s/obj-firefox/.mozbuild/build_resources.json',
                'nightly_promotion_branches': ['mozilla-central', 'mozilla-aurora'],

                # try will overwrite these
                'clone_with_purge': False,
                'clone_by_revision': False,
                'tinderbox_build_dir': None,
                'to_tinderbox_dated': True,
                'release_to_try_builds': False,
                'include_post_upload_builddir': False,
                'use_clobberer': True,

                'stage_username': 'ffxbld',
                'stage_ssh_key': 'ffxbld_rsa',
                'virtualenv_modules': [
                    'requests==2.8.1',
                    'PyHawk-with-a-single-extra-commit==0.1.5',
                    'taskcluster==0.0.26',
                ],
                'virtualenv_path': 'venv',
                #

            },
            'ConfigClass': BuildingConfig,
        }
        super(FxDesktopBuild, self).__init__(**buildscript_kwargs)

    def _pre_config_lock(self, rw_config):
        """grab buildbot props if we are running this in automation"""
        super(FxDesktopBuild, self)._pre_config_lock(rw_config)
        c = self.config
        if c['is_automation']:
            # parse buildbot config and add it to self.config
            self.info("We are running this in buildbot, grab the build props")
            self.read_buildbot_config()
            ###
            if c.get('stage_platform'):
                platform_for_log_url = c['stage_platform']
                if c.get('pgo_build'):
                    platform_for_log_url += '-pgo'
                # postrun.py uses stage_platform buildbot prop as part of the log url
                self.set_buildbot_property('stage_platform',
                                           platform_for_log_url,
                                           write_to_file=True)
            else:
                self.fatal("'stage_platform' not determined and is required in your config")


    # helpers

    def query_abs_dirs(self):
        if self.abs_dirs:
            return self.abs_dirs
        c = self.config
        abs_dirs = super(FxDesktopBuild, self).query_abs_dirs()
        if not c.get('app_ini_path'):
            self.fatal('"app_ini_path" is needed in your config for this '
                       'script.')

        dirs = {
            # BuildFactories in factory.py refer to a 'build' dir on the slave.
            # This contains all the source code/objdir to compile.  However,
            # there is already a build dir in mozharness for every mh run. The
            # 'build' that factory refers to I named: 'src' so
            # there is a seperation in mh.  for example, rather than having
            # '{mozharness_repo}/build/build/', I have '{
            # mozharness_repo}/build/src/'
            'abs_src_dir': os.path.join(abs_dirs['abs_work_dir'],
                                        'src'),
            'abs_obj_dir': os.path.join(abs_dirs['abs_work_dir'],
                                        'src',
                                        self._query_objdir()),
            'abs_tools_dir': os.path.join(abs_dirs['abs_work_dir'], 'tools'),
            'abs_app_ini_path': c['app_ini_path'] % {
                'obj_dir': os.path.join(abs_dirs['abs_work_dir'],
                                        'src',
                                        self._query_objdir())
            },
            'compare_locales_dir': os.path.join(abs_dirs['abs_work_dir'], 'compare-locales'),
        }
        abs_dirs.update(dirs)
        self.abs_dirs = abs_dirs
        return self.abs_dirs

        # Actions {{{2
        # read_buildbot_config in BuildingMixin
        # clobber in BuildingMixin -> PurgeMixin
        # if Linux config:
        # reset_mock in BuildingMixing -> MockMixin
        # setup_mock in BuildingMixing (overrides MockMixin.mock_setup)


if __name__ == '__main__':
    fx_desktop_build = FxDesktopBuild()
    fx_desktop_build.run_and_exit()



#!/usr/bin/env python
# ***** BEGIN LICENSE BLOCK *****
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
# ***** END LICENSE BLOCK *****
""" mulet_unittest.py
This allows us to run Mulet unittests

Author: Armen Zambrano Gasparnian
"""
import os
import sys

# load modules from parent dir
sys.path.insert(1, os.path.dirname(sys.path[0]))

from mozharness.base.transfer import TransferMixin
from mozharness.mozilla.gaia import GaiaMixin, gaia_config_options
from b2g_desktop_unittest import B2GDesktopTest

class MuletUnittest(B2GDesktopTest, GaiaMixin, TransferMixin):
    test_suites = ('reftest',)

    def __init__(self):
        super(MuletUnittest, self).__init__(options=gaia_config_options)

    def pull(self, **kwargs):
        GaiaMixin.pull(self, **kwargs)
        super(MuletUnittest, self).pull(**kwargs)

    def query_abs_dirs(self):
        if self.abs_dirs:
            return self.abs_dirs

        abs_dirs = super(MuletUnittest, self).query_abs_dirs()
        dirs = {}
        dirs['abs_gaia_dir'] = os.path.join(
                self.config["base_work_dir"], 'gaia')
        dirs['abs_gaia_profile'] = os.path.join(
                dirs["abs_gaia_dir"], 'profile')
        self.gaia_profile = dirs['abs_gaia_profile']
        abs_dirs.update(dirs)

        self.abs_dirs = abs_dirs
        return self.abs_dirs

    def make_gaia(self, gaia_dir):
        self.run_command(self.query_exe('make', return_type="list"),
                         cwd=gaia_dir,
                         halt_on_failure=True)

    def preflight_run_tests(self):
        # This is if we don't run all actions since we set this inside
        # of download-and-extract()
        if self.config.get("binary_path"):
            self.binary_path = self.config["binary_path"]

        if not self.binary_path:
            self.fatal("Use --binary-path as it is needed for _query_abs_dir().")

    def run_tests(self):
        """
        Run the unit test suite.
        """
        dirs = self.query_abs_dirs()

        # NOTE: We currently don't clobber the checkout or the profile
        if not os.path.isdir(dirs['abs_gaia_profile']):
            # Make the gaia profile
            self.make_gaia(dirs['abs_gaia_dir'])

        super(MuletUnittest, self).run_tests()


# main {{{1
if __name__ == '__main__':
    mulet_unittest = MuletUnittest()
    mulet_unittest.run_and_exit()



# This is a template config file for marionette test.

config = {
    # marionette options
    "test_type": "browser",
    "marionette_address": "localhost:2828",
    "test_manifest": "unit-tests.ini",

    "default_actions": [
        'clobber',
        'download-and-extract',
        'create-virtualenv',
        'install',
        'run-marionette',
    ],
    "suite_definitions": {
        "gaiatest_desktop": {
            "options": [
                "--restart",
                "--timeout=%(timeout)s",
                "--type=%(type)s",
                "--testvars=%(testvars)s",
                "--profile=%(profile)s",
                "--symbols-path=%(symbols_path)s",
                "--gecko-log=%(gecko_log)s",
                "--xml-output=%(xml_output)s",
                "--html-output=%(html_output)s",
                "--log-raw=%(raw_log_file)s",
                "--log-errorsummary=%(error_summary_file)s",
                "--binary=%(binary)s",
                "--address=%(address)s",
                "--total-chunks=%(total_chunks)s",
                "--this-chunk=%(this_chunk)s"
            ],
            "run_filename": "",
            "testsdir": ""
        },
        "gaiatest_emulator": {
            "options": [
                "--restart",
                "--timeout=%(timeout)s",
                "--type=%(type)s",
                "--testvars=%(testvars)s",
                "--profile=%(profile)s",
                "--symbols-path=%(symbols_path)s",
                "--xml-output=%(xml_output)s",
                "--html-output=%(html_output)s",
                "--log-raw=%(raw_log_file)s",
                "--log-errorsummary=%(error_summary_file)s",
                "--logcat-dir=%(logcat_dir)s",
                "--emulator=%(emulator)s",
                "--homedir=%(homedir)s"
            ],
            "run_filename": "",
            "testsdir": ""
        },
        "marionette_desktop": {
            "options": [
                "--type=%(type)s",
                "--log-raw=%(raw_log_file)s",
                "--log-errorsummary=%(error_summary_file)s",
                "--binary=%(binary)s",
                "--address=%(address)s",
                "--symbols-path=%(symbols_path)s"
            ],
            "run_filename": "",
            "testsdir": ""
        },
        "marionette_emulator": {
            "options": [
                "--type=%(type)s",
                "--log-raw=%(raw_log_file)s",
                "--log-errorsummary=%(error_summary_file)s",
                "--logcat-dir=%(logcat_dir)s",
                "--emulator=%(emulator)s",
                "--homedir=%(homedir)s",
                "--symbols-path=%(symbols_path)s"
            ],
            "run_filename": "",
            "testsdir": ""
        },
        "webapi_desktop": {
            "options": [],
            "run_filename": "",
            "testsdir": ""
        },
        "webapi_emulator": {
            "options": [
                "--type=%(type)s",
                "--log-raw=%(raw_log_file)s",
                "--log-errorsummary=%(error_summary_file)s",
                "--symbols-path=%(symbols_path)s",
                "--logcat-dir=%(logcat_dir)s",
                "--emulator=%(emulator)s",
                "--homedir=%(homedir)s"
            ],
            "run_filename": "",
            "testsdir": ""
        }
    },
}



#!/usr/bin/env python
# ***** BEGIN LICENSE BLOCK *****
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
# ***** END LICENSE BLOCK *****
"""firefox_ui_updates.py

Author: Armen Zambrano G.
        Henrik Skupin
"""
import copy
import os
import sys
import urlparse

from mozharness.base.log import FATAL, WARNING
from mozharness.base.python import PostScriptRun, PreScriptAction
from mozharness.mozilla.structuredlog import StructuredOutputParser
from mozharness.mozilla.testing.testbase import (
    TestingMixin,
    testing_config_options,
)
from mozharness.mozilla.vcstools import VCSToolsScript

# Command line arguments for firefox ui tests
firefox_ui_tests_harness_config_options = [
    [["--e10s"], {
        'dest': 'e10s',
        'action': 'store_true',
        'default': False,
        'help': 'Enable multi-process (e10s) mode when running tests.',
    }],
]

# General command line arguments for Firefox ui tests
firefox_ui_tests_config_options = [
    [['--dry-run'], {
        'dest': 'dry_run',
        'default': False,
        'help': 'Only show what was going to be tested.',
    }],
    [['--firefox-ui-branch'], {
        'dest': 'firefox_ui_branch',
        'help': 'which branch to use for firefox_ui_tests',
    }],
    [['--firefox-ui-repo'], {
        'dest': 'firefox_ui_repo',
        'default': 'https://github.com/mozilla/firefox-ui-tests.git',
        'help': 'which firefox_ui_tests repo to use',
    }],
    [['--symbols-path=SYMBOLS_PATH'], {
        'dest': 'symbols_path',
        'help': 'absolute path to directory containing breakpad '
                'symbols, or the url of a zip file containing symbols.',
    }],
] + firefox_ui_tests_harness_config_options \
    + copy.deepcopy(testing_config_options)

# Command line arguments for update tests
firefox_ui_update_harness_config_options = [
    [['--update-allow-mar-channel'], {
        'dest': 'update_allow_mar_channel',
        'help': 'Additional MAR channel to be allowed for updates, e.g. '
                '"firefox-mozilla-beta" for updating a release build to '
                'the latest beta build.',
    }],
    [['--update-channel'], {
        'dest': 'update_channel',
        'help': 'Update channel to use.',
    }],
    [['--update-direct-only'], {
        'action': 'store_true',
        'dest': 'update_direct_only',
        'help': 'Only perform a direct update.',
    }],
    [['--update-fallback-only'], {
        'action': 'store_true',
        'dest': 'update_fallback_only',
        'help': 'Only perform a fallback update.',
    }],
    [['--update-override-url'], {
        'dest': 'update_override_url',
        'help': 'Force specified URL to use for update checks.',
    }],
    [['--update-target-buildid'], {
        'dest': 'update_target_buildid',
        'help': 'Build ID of the updated build',
    }],
    [['--update-target-version'], {
        'dest': 'update_target_version',
        'help': 'Version of the updated build.',
    }],
]

firefox_ui_update_config_options = firefox_ui_update_harness_config_options \
    + copy.deepcopy(firefox_ui_tests_config_options)


class FirefoxUITests(TestingMixin, VCSToolsScript):

    # Needs to be overwritten in sub classes
    cli_script = None

    def __init__(self, config_options=None,
                 all_actions=None, default_actions=None,
                 *args, **kwargs):
        config_options = config_options or firefox_ui_tests_config_options
        actions = [
            'clobber',
            'download-and-extract',
            'create-virtualenv',
            'install',
            'run-tests',
            'uninstall',
        ]

        super(FirefoxUITests, self).__init__(
            config_options=config_options,
            all_actions=all_actions or actions,
            default_actions=default_actions or actions,
            *args, **kwargs)

        # As long as we don't run on buildbot the following properties have be set on our own
        self.binary_path = self.config.get('binary_path')
        self.installer_path = self.config.get('installer_path')
        self.installer_url = self.config.get('installer_url')
        self.test_packages_url = self.config.get('test_packages_url')
        self.test_url = self.config.get('test_url')

        self.reports = {'html': 'report.html', 'xunit': 'report.xml'}

        self.firefox_ui_repo = self.config['firefox_ui_repo']
        self.firefox_ui_branch = self.config.get('firefox_ui_branch')

        if not self.test_url and not self.test_packages_url and not self.firefox_ui_branch:
            self.fatal(
                'You must use --test-url, --test-packages-url, or --firefox-ui-branch (valid '
                'values can be found at: https://github.com/mozilla/firefox-ui-tests/branches)')

    @PreScriptAction('create-virtualenv')
    def _pre_create_virtualenv(self, action):
        dirs = self.query_abs_dirs()

        # If tests are used from common.tests.zip install every Python package
        # via the single requirements file
        if self.test_packages_url or self.test_url:
            test_install_dir = dirs.get('abs_test_install_dir',
                                        os.path.join(dirs['abs_work_dir'], 'tests'))
            requirements = os.path.join(test_install_dir,
                                        'config', 'firefox_ui_requirements.txt')
            self.register_virtualenv_module(requirements=[requirements], two_pass=True)

        # We have a non-packaged version of Firefox UI tests. So install requirements
        # and the firefox-ui-tests package separately
        # TODO - Can be removed when the github repository is no longer needed
        else:
            # Register all modules for firefox-ui-tests including all dependencies
            # as strict versions to ensure newer releases won't break something
            requirements = os.path.join(dirs.get('abs_test_install_dir',
                                                 os.path.join(dirs['abs_work_dir'], 'tests')),
                                        'requirements.txt')
            self.register_virtualenv_module(requirements=[requirements])

    def checkout(self):
        """Clone the firefox-ui-tests repository."""
        dirs = self.query_abs_dirs()

        self.vcs_checkout(
            repo=self.firefox_ui_repo,
            dest=dirs['abs_test_install_dir'],
            branch=self.firefox_ui_branch,
            vcs='gittool',
            env=self.query_env(),
        )

    def clobber(self):
        """Delete the working directory"""
        super(FirefoxUITests, self).clobber()

        # Also ensure to delete the reports directory to get rid of old files
        dirs = self.query_abs_dirs()
        self.rmtree(dirs['abs_reports_dir'], error_level=FATAL)

    def download_and_extract(self):
        """Overriding method from TestingMixin for more specific behavior.

        We use the test_packages_url command line argument to check where to get the
        harness, puppeteer, and tests from and how to set them up.

        """
        if self.test_packages_url or self.test_url:
            target_unzip_dirs = ['config/*',
                                 'firefox-ui/*',
                                 'marionette/*',
                                 'mozbase/*',
                                 'puppeteer/*',
                                 'tools/wptserve/*',
                                 ]
            super(FirefoxUITests, self).download_and_extract(target_unzip_dirs=target_unzip_dirs)

        else:
            self.checkout()
            self._download_installer()

            if self.config.get('download_symbols'):
                self._download_and_extract_symbols()

    def query_abs_dirs(self):
        if self.abs_dirs:
            return self.abs_dirs

        abs_dirs = super(FirefoxUITests, self).query_abs_dirs()
        dirs = {
            'abs_reports_dir': os.path.join(abs_dirs['base_work_dir'], 'reports'),
            'abs_test_install_dir': os.path.join(abs_dirs['abs_work_dir'], 'tests'),
        }

        for key in dirs.keys():
            if key not in abs_dirs:
                abs_dirs[key] = dirs[key]
        self.abs_dirs = abs_dirs

        return self.abs_dirs

    def query_harness_args(self, extra_harness_config_options=None):
        """Collects specific update test related command line arguments.

        Sub classes should override this method for their own specific arguments.
        """
        extra_harness_config_options = extra_harness_config_options or []
        config_options = firefox_ui_tests_harness_config_options + extra_harness_config_options

        args = []
        for option in config_options:
            dest = option[1]['dest']
            name = self.config.get(dest)

            if name:
                if type(name) is bool:
                    args.append(option[0][0])
                else:
                    args.extend([option[0][0], self.config[dest]])

        return args

    def query_minidump_stackwalk(self):
        """Download the minidump stackwalk binary.

        We can remove this whole method once we no longer need the github repository.

        """
        # If the test package is available use it
        if self.test_packages_url or self.test_url:
            return super(FirefoxUITests, self).query_minidump_stackwalk()

        # Otherwise grab the manifest file from hg.mozilla.org
        manifest_path = None

        if self.config.get('download_minidump_stackwalk'):
            tooltool_manifest = self.query_minidump_tooltool_manifest()
            url_base = 'https://hg.mozilla.org/mozilla-central/raw-file/default/testing/'

            dirs = self.query_abs_dirs()
            manifest_path = os.path.join(dirs['abs_work_dir'], 'releng.manifest')
            try:
                self.download_file(urlparse.urljoin(url_base, tooltool_manifest),
                                   manifest_path)
            except Exception as e:
                self.fatal('Download of tooltool manifest file failed: %s' % e.message)

        return super(FirefoxUITests, self).query_minidump_stackwalk(manifest=manifest_path)

    @PostScriptRun
    def copy_logs_to_upload_dir(self):
        """Overwrite this method so we also upload the other (e.g. report) files"""
        # Copy logs and report files to the upload folder
        super(FirefoxUITests, self).copy_logs_to_upload_dir()

        dirs = self.query_abs_dirs()
        self.info("Copying reports to upload dir...")
        for report in self.reports:
            self.copy_to_upload_dir(os.path.join(dirs['abs_reports_dir'], self.reports[report]),
                                    dest=os.path.join('reports', self.reports[report]),
                                    short_desc='%s log' % self.reports[report],
                                    long_desc='%s log' % self.reports[report],
                                    max_backups=self.config.get("log_max_rotate", 0),
                                    error_level=WARNING,
                                    )

    def run_test(self, binary_path, env=None, marionette_port=2828):
        """All required steps for running the tests against an installer."""
        dirs = self.query_abs_dirs()

        # Import the harness to retrieve the location of the cli scripts
        import firefox_ui_harness

        cmd = [
            self.query_python_path(),
            os.path.join(os.path.dirname(firefox_ui_harness.__file__),
                         self.cli_script),
            '--binary', binary_path,
            '--address', 'localhost:{}'.format(marionette_port),

            # Use the work dir to get temporary data stored
            '--workspace', dirs['abs_work_dir'],

            # logging options
            '--gecko-log=-',  # output from the gecko process redirected to stdout
            '--log-raw=-',  # structured log for output parser redirected to stdout

            # additional reports helpful for Jenkins and inpection via Treeherder
            '--log-html', os.path.join(dirs["abs_reports_dir"], self.reports['html']),
            '--log-xunit', os.path.join(dirs["abs_reports_dir"], self.reports['xunit']),
        ]

        # Collect all pass-through harness options to the script
        cmd.extend(self.query_harness_args())

        # Set further environment settings
        env = env or self.query_env()

        if self.symbols_url:
            cmd.extend(['--symbols-path', self.symbols_url])

        if self.query_minidump_stackwalk():
            env['MINIDUMP_STACKWALK'] = self.minidump_stackwalk_path

        parser = StructuredOutputParser(config=self.config,
                                        log_obj=self.log_obj,
                                        strict=False)

        return_code = self.run_command(cmd,
                                       cwd=dirs['abs_work_dir'],
                                       output_timeout=300,
                                       output_parser=parser,
                                       env=env)

        tbpl_status, log_level = parser.evaluate_parser(return_code)
        self.buildbot_status(tbpl_status, level=log_level)

        return return_code

    @PreScriptAction('run-tests')
    def _pre_run_tests(self, action):
        if not self.installer_path and not self.installer_url:
            self.critical('Please specify an installer via --installer-path or --installer-url.')
            sys.exit(1)

    def run_tests(self):
        """Run all the tests"""
        return self.run_test(
            binary_path=self.binary_path,
            env=self.query_env(),
        )

    def download_unzip(self, url, parent_dir, target_unzip_dirs=None, halt_on_failure=True):
        """Overwritten method from BaseScript until bug 1237706 is fixed.

        The downloaded file will always be saved to the working directory and is not getting
        deleted after extracting.

        Args:
            url (str): URL where the file to be downloaded is located.
            parent_dir (str): directory where the downloaded file will
                              be extracted to.
            target_unzip_dirs (list, optional): directories inside the zip file to extract.
                                                Defaults to `None`.
            halt_on_failure (bool, optional): whether or not to redefine the
                                              log level as `FATAL` on errors. Defaults to True.

        """
        import fnmatch
        import itertools
        import functools
        import zipfile

        def _filter_entries(namelist):
            """Filter entries of the archive based on the specified list of extract_dirs."""
            filter_partial = functools.partial(fnmatch.filter, namelist)
            for entry in itertools.chain(*map(filter_partial, target_unzip_dirs or ['*'])):
                yield entry

        dirs = self.query_abs_dirs()
        zip = self.download_file(url, parent_dir=dirs['abs_work_dir'],
                                 error_level=FATAL)

        try:
            self.info('Using ZipFile to extract {} to {}'.format(zip, parent_dir))
            with zipfile.ZipFile(zip) as bundle:
                for entry in _filter_entries(bundle.namelist()):
                    bundle.extract(entry, path=parent_dir)

                    # ZipFile doesn't preserve permissions: http://bugs.python.org/issue15795
                    fname = os.path.realpath(os.path.join(parent_dir, entry))
                    mode = bundle.getinfo(entry).external_attr >> 16 & 0x1FF
                    # Only set permissions if attributes are available.
                    if mode:
                        os.chmod(fname, mode)
        except zipfile.BadZipfile as e:
            self.log('%s (%s)' % (e.message, zip),
                     level=FATAL, exit_code=2)


class FirefoxUIFunctionalTests(FirefoxUITests):

    cli_script = 'cli_functional.py'


class FirefoxUIUpdateTests(FirefoxUITests):

    cli_script = 'cli_update.py'

    def __init__(self, config_options=None, *args, **kwargs):
        config_options = config_options or firefox_ui_update_config_options

        super(FirefoxUIUpdateTests, self).__init__(
            config_options=config_options,
            *args, **kwargs
        )

    def query_harness_args(self):
        """Collects specific update test related command line arguments."""
        return super(FirefoxUIUpdateTests, self).query_harness_args(
            firefox_ui_update_harness_config_options)



""" googleplay.py

    The way to get the API access is to
      1) login in in the Google play admin
      2) Settings
      3) API Access
      4) go in the Google Developers Console
      5) Create "New client ID"
         or download the p12 key (it should remain
         super private)
      6) Move the file in this directory with the name
         'key.p12' or use the --credentials option
"""

import httplib2

from apiclient.discovery import build
from oauth2client import client


# GooglePlayMixin {{{1
class GooglePlayMixin(object):

    def connect_to_play(self):
        """ Connect to the google play interface
        """
        # Load the key in PKCS 12 format that you downloaded from the
        # Google APIs Console when you created your Service account.
        f = file(self.config["google_play_credentials_file"], 'rb')
        key = f.read()
        f.close()

        # Create an httplib2.Http object to handle our HTTP requests an
        # authorize it with the Credentials. Note that the first parameter,
        # service_account_name, is the Email address created for the Service
        # account. It must be the email address associated with the key that
        # was created.
        scope = 'https://www.googleapis.com/auth/androidpublisher'
        credentials = client.SignedJwtAssertionCredentials(
            self.config["service_account"],
            key,
            scope)
        http = httplib2.Http()
        http = credentials.authorize(http)

        service = build('androidpublisher', 'v2', http=http)

        return service



LIVE_B2G_BRANCHES = {
    "mozilla-b2g37_v2_2": {
        "gaia_branch": "v2.2",
        "tag_name": "B2G_2_2_%(DATE)s_MERGEDAY",
    },
    "mozilla-b2g37_v2_2r": {
        "gaia_branch": "v2.2r",
        "tag_name": "B2G_2_2r_%(DATE)s_MERGEDAY",
    },
}

config = {
    "log_name": "b2g_tag",

    "gaia_mapper_base_url": "http://cruncher/mapper/gaia/git",
    "gaia_url": "git@github.com:mozilla-b2g/gaia.git",
    "hg_base_pull_url": "https://hg.mozilla.org/releases",
    "hg_base_push_url": "ssh://hg.mozilla.org/releases",
    "b2g_branches": LIVE_B2G_BRANCHES,

    # Disallow sharing, since we want pristine .hg directories.
    "vcs_share_base": None,
    "hg_share_base": None,

    # any hg command line options
    "exes": {
        "hg": [
            "hg", "--config",
            "hostfingerprints.hg.mozilla.org=af:27:b9:34:47:4e:e5:98:01:f6:83:2b:51:c9:aa:d8:df:fb:1a:27",
        ],
    }
}



import os

config = {
    'default_actions': [
        'clobber',
        'clone-tools',
        'checkout-sources',
        'setup-mock',
        'build',
        'upload-files',
        'sendchange',
        'check-test',
        'update',  # may or may not happen based on query_is_nightly()
    ],
    'stage_platform': 'linux64',
    'stage_product': 'b2g',
    'enable_signing': False,
    'enable_talos_sendchange': False,
    'enable_unittest_sendchange': False,
    'enable_count_ctors': False,
    'objdir': 'obj-horizon',
    #### 64 bit build specific #####
    'env': {
        'MOZBUILD_STATE_PATH': os.path.join(os.getcwd(), '.mozbuild'),
        'MOZ_AUTOMATION': '1',
        'HG_SHARE_BASE_DIR': '/builds/hg-shared',
        'MOZ_OBJDIR': 'obj-horizon',
        'TINDERBOX_OUTPUT': '1',
        'CCACHE_DIR': '/builds/ccache',
        'CCACHE_COMPRESS': '1',
        'CCACHE_UMASK': '002',
        'LC_ALL': 'C',
        ## 64 bit specific
        'PATH': '/tools/buildbot/bin:/usr/local/bin:/usr/lib64/ccache:/bin:\
/usr/bin:/usr/local/sbin:/usr/sbin:/sbin:/tools/git/bin:/tools/python27/bin:\
/tools/python27-mercurial/bin:/home/cltbld/bin',
        "SYMBOL_SERVER_HOST": "%(symbol_server_host)s",
        "SYMBOL_SERVER_SSH_KEY": "/home/mock_mozilla/.ssh/ffxbld_rsa",
        "SYMBOL_SERVER_USER": "ffxbld",
        "SYMBOL_SERVER_PATH": "/mnt/netapp/breakpad/symbols_ffx",
    },
    'src_mozconfig': 'b2g/graphene/config/horizon-mozconfigs/linux64/nightly',
    'balrog_platform': 'linux64',
    #######################
}



config = {
    "log_name": "bump_esr38",
    "version_files": [
        {"file": "browser/config/version.txt"},
        # TODO: enable this for esr45
        # {"file": "browser/config/version_display.txt"},
        {"file": "config/milestone.txt"},
    ],
    "repo": {
        "repo": "https://hg.mozilla.org/releases/mozilla-esr38",
        "revision": "default",
        "dest": "mozilla-esr38",
        "vcs": "hg",
    },
    "push_dest": "ssh://hg.mozilla.org/releases/mozilla-esr38",
    "ignore_no_changes": True,
    "ssh_user": "ffxbld",
    "ssh_key": "~/.ssh/ffxbld_rsa",
}



"""
Module for performing gaia-specific tasks
"""

import json
import os
import re

from mozharness.base.errors import HgErrorList, BaseErrorList, ZipErrorList
from mozharness.base.log import ERROR, FATAL

gaia_config_options = [
    [["--gaia-dir"],
    {"action": "store",
     "dest": "gaia_dir",
     "help": "directory where gaia repo should be cloned"
     }],
    [["--gaia-repo"],
    {"action": "store",
     "dest": "gaia_repo",
     "help": "url of gaia repo to clone"
    }],
    [["--gaia-branch"],
    {"action": "store",
     "dest": "gaia_branch",
     "default": "default",
     "help": "branch of gaia repo to clone"
    }],
]

class GaiaMixin(object):

    npm_error_list = BaseErrorList + [
        {'substr': r'''npm ERR! Error:''', 'level': ERROR}
    ]

    # This requires self to inherit a VCSMixin.
    def clone_gaia(self, dest, repo, use_gaia_json=False):
        """
        Clones an hg mirror of gaia.

        repo: a dict containing 'repo_path', 'revision', and optionally
              'branch' parameters
        use_gaia_json: if True, the repo parameter is used to retrieve
              a gaia.json file from a gecko repo, which in turn is used to
              clone gaia; if False, repo represents a gaia repo to clone.
        """

        repo_path = repo.get('repo_path')
        revision = repo.get('revision')
        branch = repo.get('branch')
        gaia_json_path = self.config.get("gaia_json_path", "{repo_path}/raw-file/{revision}/b2g/config/gaia.json")
        git = False
        pr_git_revision = None
        pr_remote = None

        self.info('dest: %s' % dest)

        if use_gaia_json:
            url = gaia_json_path.format(
                repo_path=repo_path,
                revision=revision)
            contents = self.retry(self.load_json_from_url, args=(url,))
            if contents.get('git') and contents['git'].get('remote'):
                git = True
                remote = contents['git']['remote']
                branch = contents['git'].get('branch')
                revision = contents['git'].get('git_revision')
                pr_git_revision = contents['git'].get('pr_git_revision')
                pr_remote = contents['git'].get('pr_remote')
                if pr_remote or pr_git_revision:
                    if not (pr_remote and pr_git_revision):
                        self.fatal('Pull request mode requres rev *and* remote')
                if not (branch or revision):
                    self.fatal('Must specify branch or revision for git repo')
            elif contents.get('repo_path') and contents.get('revision'):
                repo_path = 'https://hg.mozilla.org/%s' % contents['repo_path']
                revision = contents['revision']
                branch = None

        if git:
            git_cmd = self.query_exe('git')
            needs_clobber = True

            # For pull requests, we only want to clobber when we can't find the
            # two exact commit ids that we'll be working with.  As long as we
            # have those two commits, we don't care about the rest of the repo
            def has_needed_commit(commit, fatal=False):
                cmd = [git_cmd, 'rev-parse', '--quiet', '--verify', '%s^{commit}' % commit]
                rc = self.run_command(cmd, cwd=dest, halt_on_failure=False, success_codes=[1,0])
                if rc != 0:
                    return False
                return True

            if not pr_remote and os.path.exists(dest) and os.path.exists(os.path.join(dest, '.git')):
                cmd = [git_cmd, 'remote', '-v']
                output = self.get_output_from_command(cmd, cwd=dest)
                for line in output:
                    if remote in line:
                        needs_clobber = False


            # We want to do some cleanup logic differently for pull requests
            if pr_git_revision and pr_remote:
                needs_clobber = False
                if os.path.exists(dest) and os.path.exists(os.path.join(dest, '.git')):
                    cmd = [git_cmd, 'clean', '-f', '-f', '-x', '-d']
                    self.run_command(cmd, cwd=dest, halt_on_failure=True,
                                     fatal_exit_code=3)
                    if not has_needed_commit(revision):
                        cmd = [git_cmd, 'fetch', 'origin']
                        self.run_command(cmd, cwd=dest, halt_on_failure=True,
                                         fatal_exit_code=3)
                    if not has_needed_commit(revision):
                        self.warn('Repository does not contain required revisions, clobbering')
                        needs_clobber = True

            # In pull request mode, we don't want to clone because we're satisfied
            # that the base directory is good enough because
            needs_clone = True
            if pr_git_revision and pr_remote:
                if os.path.exists(dest):
                    if os.path.exists(os.path.join(dest, '.git')):
                        needs_clone = False
                    else:
                        needs_clobber = True

            if needs_clobber:
                self.rmtree(dest)
                needs_clone = True

            if needs_clone:
                # git clone
                cmd = [git_cmd,
                       'clone',
                       remote]
                self.run_command(cmd,
                                 cwd=os.path.dirname(dest),
                                 output_timeout=1760,
                                 halt_on_failure=True,
                                 fatal_exit_code=3)

            self.run_command([git_cmd, 'status'], cwd=dest);


            # handle pull request magic
            if pr_git_revision and pr_remote:
                # Optimization opportunity: instead of fetching all remote references,
                # pull only the single commit.  I don't know how to right now

                # If the 'other' remote exists, get rid of it
                cmd = [git_cmd, 'remote']
                output = self.get_output_from_command(cmd, cwd=dest)
                for line in output.split('\n'):
                  if 'other' in line:
                    cmd = [git_cmd, 'remote', 'rm', 'other']
                    self.run_command(cmd, cwd=dest, halt_on_failure=True,
                                     fatal_exit_code=3)
                    break;
                # Set the correct remote
                cmd = [git_cmd, 'remote', 'add', 'other', pr_remote]
                self.run_command(cmd, cwd=dest, halt_on_failure=True,
                                 fatal_exit_code=3)
                if not has_needed_commit(pr_git_revision):
                    cmd = [git_cmd, 'fetch', 'other']
                    self.run_command(cmd, cwd=dest, halt_on_failure=True,
                                     fatal_exit_code=3)
                if not has_needed_commit(pr_git_revision):
                    self.fatal('Missing the Pull Request target revision')

                # With these environment variables we should have deterministic
                # merge commit identifiers
                self.info('If you want to prove that this merge commit is the same')
                self.info('you get, use this environment while doing the merge')
                env = {
                  'GIT_COMMITTER_DATE': "Wed Feb 16 14:00 2037 +0100",
                  'GIT_AUTHOR_DATE': "Wed Feb 16 14:00 2037 +0100",
                  'GIT_AUTHOR_NAME': 'automation',
                  'GIT_AUTHOR_EMAIL': 'auto@mati.on',
                  'GIT_COMMITTER_NAME': 'automation',
                  'GIT_COMMITTER_EMAIL': 'auto@mati.on'
                }
                cmd = [git_cmd, 'reset', '--hard', revision or branch]
                self.run_command(cmd, cwd=dest, halt_on_failure=True,
                                 fatal_exit_code=3)
                cmd = [git_cmd, 'clean', '-f', '-f', '-x', '-d']
                self.run_command(cmd, cwd=dest, halt_on_failure=True,
                                 fatal_exit_code=3)
                cmd = [git_cmd, 'merge', '--no-ff', pr_git_revision]
                self.run_command(cmd, cwd=dest, env=env, halt_on_failure=True,
                                 fatal_exit_code=3)
                # So that people can verify that their merge commit is identical
                cmd = [git_cmd, 'rev-parse', 'HEAD']
                self.run_command(cmd, cwd=dest, halt_on_failure=True,
                                 fatal_exit_code=3)

            else:
                # checkout git branch
                cmd = [git_cmd,
                       'checkout',
                       revision or branch]
                self.run_command(cmd, cwd=dest, halt_on_failure=True,
                                 fatal_exit_code=3)


            # verify
            for cmd in ([git_cmd, 'rev-parse', 'HEAD'], [git_cmd, 'branch']):
                self.run_command(cmd, cwd=dest, halt_on_failure=True,
                                 fatal_exit_code=3)

        else:
            # purge the repo if it already exists
            if os.path.exists(dest):
                if os.path.exists(os.path.join(dest, '.hg')):
                    # this is an hg dir, so do an hg clone
                    cmd = [self.query_exe('hg'),
                           '--config',
                           'extensions.purge=',
                           'purge']
                    if self.run_command(cmd, cwd=dest, error_list=HgErrorList):
                        self.fatal("Unable to purge %s!" % dest)
                else:
                    # there's something here, but it isn't hg; just delete it
                    self.rmtree(dest)

            repo = {
                'repo': repo_path,
                'revision': revision,
                'branch': branch,
                'dest': dest,
            }

            self.vcs_checkout_repos([repo], parent_dir=os.path.dirname(dest))

    def preflight_pull(self):
        if not self.buildbot_config:
            if not self.config.get('gaia_repo'):
                # gaia_branch by default is set to default
                self.fatal("You're trying to run this outside of buildbot, " \
                    "therefore, you need to specify --gaia-repo.")
            if not self.config.get('gaia_branch'):
                # gaia_branch by default is set to default
                self.fatal("You're trying to run this outside of buildbot, " \
                    "therefore, you need to specify --gaia-branch.")

    def extract_xre(self, xre_url, xre_path=None, parent_dir=None):
        if not xre_path:
            xre_path = self.config.get('xre_path')
        xulrunner_bin = os.path.join(parent_dir,
                                     xre_path,
                                     'bin', 'xpcshell')
        if os.access(xulrunner_bin, os.F_OK):
            return

        filename = self.download_file(xre_url, parent_dir=parent_dir)
        m = re.search('\.tar\.(bz2|gz)$', filename)
        if m:
            # a xulrunner archive, which has a top-level 'xulrunner-sdk' dir
            command = self.query_exe('tar', return_type='list')
            tar_cmd = "jxf"
            if m.group(1) == "gz":
                tar_cmd = "zxf"
            command.extend([tar_cmd, filename])
            self.run_command(command,
                             cwd=parent_dir,
                             error_list=TarErrorList,
                             halt_on_failure=True,
                             fatal_exit_code=3)
        else:
            # a tooltool xre.zip
            command = self.query_exe('unzip', return_type='list')
            command.extend(['-q', '-o', filename])
            # Gaia assumes that xpcshell is in a 'xulrunner-sdk' dir, but
            # xre.zip doesn't have a top-level directory name, so we'll
            # create it.
            parent_dir = os.path.join(parent_dir,
                                      self.config.get('xre_path'))
            if not os.access(parent_dir, os.F_OK):
                self.mkdir_p(parent_dir, error_level=FATAL)
            self.run_command(command,
                             cwd=parent_dir,
                             error_list=ZipErrorList,
                             halt_on_failure=True,
                             fatal_exit_code=3)

    def pull(self, **kwargs):
        '''
        Two ways of using this function:
        - The user specifies --gaia-repo or in a config file
        - The buildbot propeties exist and we query the gaia json url
          for the current gecko tree
        '''
        dirs = self.query_abs_dirs()
        dest = dirs['abs_gaia_dir']
        repo = {}

        if self.buildbot_config is not None:
            # get gaia commit via hgweb (gecko's gaia.json)
            repo = {
                'revision': self.buildbot_config['properties']['revision'],
                'repo_path': 'https://hg.mozilla.org/%s' % self.buildbot_config['properties']['repo_path'],
                'branch': None,
            }
        else:
            repo = {
                'repo_path': self.config['gaia_repo'],
                'revision': 'default',
                'branch': self.config['gaia_branch']
            }

        self.clone_gaia(dest, repo,
                        use_gaia_json=self.buildbot_config is not None)

    def make_gaia(self, gaia_dir, xre_dir, debug=False, noftu=True,
                  xre_url=None, build_config_path=None):
        if xre_url:
            self.extract_xre(xre_url, xre_path=xre_dir, parent_dir=gaia_dir)

        env = {'DEBUG': '1' if debug else '0',
               'NOFTU': '1' if noftu else '0',
               'DESKTOP': '0',
               'DESKTOP_SHIMS': '1',
               'USE_LOCAL_XULRUNNER_SDK': '1',
               'XULRUNNER_DIRECTORY': xre_dir
              }

        # if tbpl_build_config.json exists, load it
        if build_config_path:
            if os.path.exists(build_config_path):
                with self.opened(build_config_path) as (f, err):
                    if err:
                        self.fatal("Error while reading %s, aborting" %
                                   build_config_path)
                    else:
                        contents = f.read()
                        config = json.loads(contents)
                        env.update(config.get('env', {}))

        self.info('Sending environment as make vars because of bug 1028816')

        cmd = self.query_exe('make', return_type="list")
        for key, value in env.iteritems():
            cmd.append('%s=%s' % (key, value))
        self.run_command(cmd,
                         cwd=gaia_dir,
                         halt_on_failure=True)

    def make_node_modules(self):

        dirs = self.query_abs_dirs()

        def cleanup_node_modules():
            node_dir = os.path.join(dirs['abs_gaia_dir'], 'node_modules')
            self.rmtree(node_dir)

        self.run_command(['npm', 'cache', 'clean'])

        # get the node modules first from the node module cache, if this fails it will 
        # install the node modules from npmjs.org.
        cmd = ['taskcluster-npm-cache-get',
               '--namespace',
               'gaia.npm_cache',
               'package.json']
        kwargs = {
            'output_timeout': 300
        }
        code = self.retry(self.run_command, attempts=3, good_statuses=(0,),
                          args=[cmd, dirs['abs_gaia_dir']], cleanup=cleanup_node_modules, kwargs=kwargs)
        if code:
            self.fatal('Error occurred fetching node modules from cache',
                       exit_code=code)

        # run 'make node_modules' second, so we can separately handle
        # errors that occur here
        cmd = ['make',
               'node_modules',
               'NODE_MODULES_SRC=npm-cache',
               'VIRTUALENV_EXISTS=1']
        kwargs = {
            'output_timeout': 300,
            'error_list': self.npm_error_list
        }
        code = self.retry(self.run_command, attempts=3, good_statuses=(0,),
                          args=[cmd, dirs['abs_gaia_dir']], cleanup=cleanup_node_modules, kwargs=kwargs)
        if code:
            # Dump npm-debug.log, if it exists
            npm_debug = os.path.join(dirs['abs_gaia_dir'], 'npm-debug.log')
            if os.access(npm_debug, os.F_OK):
                self.info('dumping npm-debug.log')
                self.run_command(['cat', npm_debug])
            else:
                self.info('npm-debug.log doesn\'t exist, not dumping')
            self.fatal('Errors during \'npm install\'', exit_code=code)

        cmd = ['make',
               'update-common']
        kwargs = {
            'output_timeout': 300
        }
        code = self.retry(self.run_command, attempts=3, good_statuses=(0,),
                          args=[cmd, dirs['abs_gaia_dir']], kwargs=kwargs)
        if code:
            self.fatal('Errors during make update-common')

    def node_setup(self):
        """
        Set up environment for node-based Gaia tests.
        """
        dirs = self.query_abs_dirs()

        # Set-up node modules for project first. They must be older than 
        # the b2g build we copy into place. Otherwise the b2g build will
        # be considered an out of date target and we don't want that! It
        # can cause our desired b2g-desktop build to be overwritten!
        self.make_node_modules()

        # Copy the b2g desktop we built to the gaia directory so that it
        # gets used by the marionette-js-runner.
        b2g_dest = os.path.join(dirs['abs_gaia_dir'], 'b2g')
        self.copytree(
            os.path.join(os.path.dirname(self.binary_path)),
            b2g_dest,
            overwrite='clobber'
        )
        # Ensure modified time is more recent than node_modules!
        self.run_command(['touch', '-c', b2g_dest])




#!/usr/bin/env python
# ***** BEGIN LICENSE BLOCK *****
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
# ***** END LICENSE BLOCK *****

import sys, os, re, multiprocessing
import subprocess, pprint, time, datetime, getpass, platform
from ftplib import FTP

# load modules from parent dir
sys.path.insert(1, os.path.dirname(sys.path[0]))

from mozharness.base.script import BaseScript
from mozharness.mozilla.purge import PurgeMixin

CONFIG_INI = {
    "avd.ini.encoding": "ISO-8859-1",
    "hw.dPad": "no",
    "hw.lcd.density": "160",
    "sdcard.size": "500M",
    "hw.cpu.arch": "arm",
    "hw.gpu.enabled": "yes",
    "hw.device.hash": "-671137758",
    "hw.camera.back": "none",
    "disk.dataPartition.size": "600M",
    "skin.path": "1024x816",
    "skin.dynamic": "yes",
    "hw.keyboard.lid": "yes",
    "hw.keyboard": "yes",
    "hw.ramSize": "1024",
    "hw.device.manufacturer": "User",
    "hw.sdCard": "yes",
    "hw.mainKeys": "no",
    "hw.accelerometer": "yes",
    "skin.name": "1024x816",
    "hw.trackBall": "no",
    "hw.device.name": "mozilla-device",
    "hw.battery": "yes",
    "hw.sensors.proximity": "yes",
    "hw.sensors.orientation": "yes",
    "hw.audioInput": "yes",
    "hw.gps": "yes",
    "vm.heapSize": "64",
    }


MOZILLA_DEVICE_DEFINITION = """
<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<d:devices xmlns:d="http://schemas.android.com/sdk/devices/1" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
  <d:device>
    <d:name>mozilla-device</d:name>
    <d:manufacturer>User</d:manufacturer>
    <d:meta/>
    <d:hardware>
      <d:screen>
        <d:screen-size>large</d:screen-size>
        <d:diagonal-length>7.00</d:diagonal-length>
        <d:pixel-density>mdpi</d:pixel-density>
        <d:screen-ratio>long</d:screen-ratio>
        <d:dimensions>
          <d:x-dimension>1024</d:x-dimension>
          <d:y-dimension>816</d:y-dimension>
        </d:dimensions>
        <d:xdpi>187.05</d:xdpi>
        <d:ydpi>187.05</d:ydpi>
        <d:touch>
          <d:multitouch>jazz-hands</d:multitouch>
          <d:mechanism>finger</d:mechanism>
          <d:screen-type>capacitive</d:screen-type>
        </d:touch>
      </d:screen>
      <d:networking>
Bluetooth
Wifi
NFC</d:networking>
      <d:sensors>
Accelerometer
Barometer
Compass
GPS
Gyroscope
LightSensor
ProximitySensor</d:sensors>
      <d:mic>true</d:mic>
      <d:camera>
        <d:location>back</d:location>
        <d:autofocus>true</d:autofocus>
        <d:flash>true</d:flash>
      </d:camera>
      <d:keyboard>qwerty</d:keyboard>
      <d:nav>nonav</d:nav>
      <d:ram unit="GiB">1</d:ram>
      <d:buttons>soft</d:buttons>
      <d:internal-storage unit="GiB">
4</d:internal-storage>
      <d:removable-storage unit="TiB"/>
      <d:cpu>Generic CPU</d:cpu>
      <d:gpu>Generic GPU</d:gpu>
      <d:abi>
armeabi
armeabi-v7a
x86
mips</d:abi>
      <d:dock/>
      <d:power-type>battery</d:power-type>
    </d:hardware>
    <d:software>
      <d:api-level>-</d:api-level>
      <d:live-wallpaper-support>true</d:live-wallpaper-support>
      <d:bluetooth-profiles/>
      <d:gl-version>2.0</d:gl-version>
      <d:gl-extensions/>
      <d:status-bar>false</d:status-bar>
    </d:software>
    <d:state default="true" name="Portrait">
      <d:description>The device in portrait orientation</d:description>
      <d:screen-orientation>port</d:screen-orientation>
      <d:keyboard-state>keyshidden</d:keyboard-state>
      <d:nav-state>navhidden</d:nav-state>
    </d:state>
    <d:state name="Landscape">
      <d:description>The device in landscape orientation</d:description>
      <d:screen-orientation>land</d:screen-orientation>
      <d:keyboard-state>keyshidden</d:keyboard-state>
      <d:nav-state>navhidden</d:nav-state>
    </d:state>
    <d:state name="Portrait with keyboard">
      <d:description>The device in portrait orientation with a keyboard open</d:description>
      <d:screen-orientation>land</d:screen-orientation>
      <d:keyboard-state>keysexposed</d:keyboard-state>
      <d:nav-state>navhidden</d:nav-state>
    </d:state>
    <d:state name="Landscape with keyboard">
      <d:description>The device in landscape orientation with a keyboard open</d:description>
      <d:screen-orientation>land</d:screen-orientation>
      <d:keyboard-state>keysexposed</d:keyboard-state>
      <d:nav-state>navhidden</d:nav-state>
    </d:state>
  </d:device>
</d:devices>
"""

def sniff_host_arch():
    host_arch = 'x86_64'
    if platform.architecture()[0] == '32bit':
        host_arch = 'x86'
    return host_arch

class EmulatorBuild(BaseScript, PurgeMixin):

    config_options = [
        [["--host-arch"], {
            "dest": "host_arch",
            "help": "architecture of the host the emulator will run on (x86, x86_64; default autodetected)",
        }],
        [["--target-arch"], {
            "dest": "target_arch",
            "help": "architecture of the target the emulator will emulate (armv5te, armv7a, x86; default armv7a)",
        }],
        [["--android-version"], {
            "dest": "android_version",
            "help": "android version to build (eg. 2.3.7, 4.0, 4.3.1, gingerbread, ics, jb; default gingerbread)",
        }],
        [["--android-tag"], {
            "dest": "android_tag",
            "help": "android tag to check out (eg. android-2.3.7_r1; default inferred from --android-version)",
        }],
        [["--patch"], {
            "dest": "patch",
            "help": "'dir=url' comma-separated list of patches to apply to AOSP before building (eg. development=http://foo.com/bar.patch; default inferred)",
        }],
        [["--android-apilevel"], {
            "dest": "android_apilevel",
            "help": "android API-level to build AVD for (eg. 10, 14, 18; default inferred from --android-version)",
        }],
        [["--android-url"], {
            "dest": "android_url",
            "help": "where to fetch AOSP from, default https://android.googlesource.com/platform/manifest",
        }],
        [["--ndk-version"], {
            "dest": "ndk_version",
            "help": "version of the NDK to fetch, default r9",
        }],
        [["--install-android-dir"], {
            "dest": "install_android_dir",
            "help": "location bundled AVDs will be unpacked to, default /home/cltbld/.android"
        }],
        [["--avd-count"], {
            "dest": "avd_count",
            "help": "number of AVDs to build, default 4"
        }],
        [["--jdk"], {
            "dest": "jdk",
            "help": "which jdk to use (sun or openjdk; default sun)"
        }]
    ]

    def __init__(self, require_config_file=False):

        BaseScript.__init__(self,
                            config_options=self.config_options,
                            all_actions=[
                                'clobber',
                                'apt-get-dependencies',
                                'download-aosp',
                                'download-kernel',
                                'download-ndk',
                                'download-test-binaries',
                                'checkout-orangutan',
                                'patch-aosp',
                                'build-aosp',
                                'build-kernel',
                                'build-orangutan-su',
                                'make-base-avd',
                                'customize-avd',
                                'clone-customized-avd',
                                'bundle-avds',
                                'bundle-emulators'
                            ],
                            default_actions=[
                                'apt-get-dependencies',
                                'download-aosp',
                                'download-kernel',
                                'download-ndk',
                                'download-test-binaries',
                                'checkout-orangutan',
                                'patch-aosp',
                                'build-aosp',
                                'build-kernel',
                                'build-orangutan-su',
                                'make-base-avd',
                                'customize-avd',
                                'clone-customized-avd',
                                'bundle-avds',
                                'bundle-emulators'
                            ],
                            require_config_file=require_config_file,
                            # Default configuration
                            config={
                                'host_arch': sniff_host_arch(),
                                'target_arch': 'armv7a',
                                'android_version': 'gingerbread',
                                'android_tag': 'inferred',
                                'patch': 'inferred',
                                'android_apilevel': 'inferred',
                                'work_dir': 'android_emulator_build',
                                'android_url': 'https://android.googlesource.com/platform/manifest',
                                'ndk_version': 'r9',
                                'install_android_dir': '/home/cltbld/.android',
                                'avd_count': '4',
                                'jdk': 'sun'
                            })

        if platform.system() != "Linux":
            self.fatal("this script only works on (ubuntu) linux")

        if platform.dist() != ('Ubuntu', '12.04', 'precise'):
            self.fatal("this script only works on ubuntu 12.04 precise")

        if not (platform.machine() in ['i386', 'i486', 'i586', 'i686', 'x86_64']):
            self.fatal("this script only works on x86 and x86_64")

        self.tag = self.config['android_tag']
        if self.tag == 'inferred':
            self.tag = self.select_android_tag(self.config['android_version'])

        self.patches = self.config['patch']
        if self.patches == 'inferred':
            self.patches = self.select_patches(self.tag)
        else:
            self.patches = [x.split('=') for x in self.patches.split(',')]

        self.apilevel = self.config['android_apilevel']
        if self.apilevel == 'inferred':
            self.apilevel = self.android_apilevel(self.tag)

        self.workdir = os.path.abspath(self.config['work_dir'])
        self.bindir = os.path.join(self.workdir, "bin")
        self.aospdir = os.path.join(self.workdir, "aosp")
        self.goldfishdir = os.path.join(self.workdir, "goldfish")
        self.ndkdir = os.path.join(self.workdir, "android-ndk-" + self.config['ndk_version'])
        self.ncores = multiprocessing.cpu_count()
        self.androiddir = os.path.join(self.workdir, ".android")
        self.avddir = os.path.join(self.androiddir, "avd")
        self.aosphostdir = os.path.join(self.aospdir, "out/host/linux-x86")
        self.aosphostbindir = os.path.join(self.aosphostdir, "bin")
        self.aospprodoutdir = os.path.join(self.aospdir, "out/target/product/generic")
        self.emu = os.path.join(self.aosphostbindir, "emulator")
        self.adb = os.path.join(self.aosphostbindir, "adb")
        self.navds = int(self.config['avd_count'])

    def apt_get_dependencies(self):
        jdk = "oracle-java6-installer"
        if self.config['jdk'] == "openjdk":
            jdk = "openjdk-6-jdk"
        self.apt_update()
        self.apt_get(["python-software-properties"])
        for repo in [ "ppa:webupd8team/java",
                      "deb http://archive.canonical.com/ precise partner",
                      "deb http://us.archive.ubuntu.com/ubuntu/ precise universe",
                      "deb http://us.archive.ubuntu.com/ubuntu/ precise-updates universe",
                      "deb http://us.archive.ubuntu.com/ubuntu/ precise-backports main restricted universe multiverse",
                      "deb http://security.ubuntu.com/ubuntu precise-security universe" ]:
            self.apt_add_repo(repo)
        self.apt_update()
        self.apt_get(["python-software-properties"])
        for pkgset in [[jdk],
                       ["libglw1-mesa"],
                       ["git", "gnupg", "flex", "bison", "gperf", "zip", "curl"],
                       ["mingw32", "tofrodos"],
                       ["build-essential", "gcc-4.4-multilib", "g++-4.4-multilib"],
                       ["python-markdown", "libxml2-utils", "xsltproc"],
                       ["x11proto-core-dev:i386", "libx11-dev:i386", "libreadline6-dev:i386",
                        "libgl1-mesa-glx:i386", "libgl1-mesa-dev:i386", "mesa-common-dev:i386",
                        "libxext-dev:i386"],
                       ["libc6-dev:i386", "libncurses5-dev:i386", "zlib1g-dev:i386"],
                       ["zlib1g-dev"],
                       ["gcc", "g++"]]:
            self.apt_get(pkgset)


    def download_aosp(self):
        self.download_file("http://commondatastorage.googleapis.com/git-repo-downloads/repo",
                           file_name="repo", parent_dir=self.bindir)
        repo = os.path.join(self.bindir, "repo")
        self.chmod(repo, 0755)

        self.mkdir_p(self.aospdir)
        self.run_command([repo, "init", "-u", self.config['android_url'],
                          "-b", self.tag],
                         cwd=self.aospdir,
                         halt_on_failure=True)
        self.run_command([repo, "sync", "-j", str(self.ncores)],
                         cwd=self.aospdir,
                         halt_on_failure=True)

        # Note: AOSP upstream tagged a version of the emulator with 2.3.x that will not
        # actually boot gingerbread images. Quoting their explanation for posterity here, from
        # http://source.android.com/source/known-issues.html:
        #
        #     Symptom: The emulator built directly from the gingerbread branch doesn't start and
        #     stays stuck on a black screen.
        #
        #     Cause: The gingerbread branch uses version R7 of the emulator, which doesn't have all
        #     the features necessary to run recent versions of gingerbread.
        #
        #     Fix: Use version R12 of the emulator, and a newer kernel that matches those tools. No
        #     need to do a clean build.
        #
        if self.tag.startswith("android-2.3"):
            self.info("updating QEMU sub-repository to R12")
            self.info("to compensate for known 2.3.x-incompatible R7 in-tree")
            self.run_command([repo, "forall", "platform/external/qemu",
                              "-c", "git", "checkout", "aosp/tools_r12"],
                             cwd=self.aospdir,
                             halt_on_failure=True)

        # For 2.3.x you _probably_ want to use the symbolic tag 'gingerbread' because it is a
        # post-release development branch on which they have backported the GL emulation that
        # fennec relies on. In order to support that, we need to update qemu to R17.
        if self.tag == 'gingerbread':
            self.info("updating QEMU sub-repository to R17")
            self.info("to acquire machinery needed for GL emulation backport")
            self.run_command([repo, "forall", "platform/external/qemu",
                              "-c", "git", "checkout", "aosp/tools_r17"],
                             cwd=self.aospdir,
                             halt_on_failure=True)

    def download_kernel(self):
        self.mkdir_p(self.workdir)
        self.run_command(["git", "clone",
                          "https://android.googlesource.com/kernel/goldfish.git"],
                         cwd=self.workdir,
                         halt_on_failure=True)
        self.run_command(["git", "checkout",
                          "origin/android-goldfish-2.6.29"],
                         cwd=self.goldfishdir,
                         halt_on_failure=True)


    def download_ndk(self):
        ndk = "android-ndk-%s-linux-%s.tar.bz2" % (self.config['ndk_version'], self.config['host_arch'])
        self.download_file("http://dl.google.com/android/ndk/" + ndk,
                           file_name=ndk, parent_dir=self.workdir)
        self.run_command(["tar", "-xjf", ndk],
                         cwd=self.workdir, halt_on_failure=True)


    def download_test_binaries(self):
            lines = []
            zipname = None
            host = "ftp.mozilla.org"
            path = None
            if self.is_arm_target():
                path = 'pub/mobile/nightly/latest-mozilla-central-android'
            else:
                path = 'pub/mobile/nightly/latest-mozilla-central-android-x86'
            ftp = FTP(host)
            ftp.login()
            ftp.cwd(path)
            ftp.retrlines('NLST', lambda x: lines.append(x.strip()))
            for line in lines:
                if line.endswith("tests.zip"):
                    zipname = line
                    break
            if zipname == None:
                self.fatal("unable to find *tests.zip at ftp://%s/%s" % (host,path))
            url = "ftp://%s/%s/%s" % (host,path,zipname)
            self.download_file(url, file_name=zipname, parent_dir=self.workdir)
            self.run_command(["unzip", zipname,
                              "bin/sutAgentAndroid.apk",
                              "bin/Watcher.apk"],
                             cwd=self.workdir,
                             halt_on_failure=True)


    def checkout_orangutan(self):
        self.mkdir_p(self.workdir)
        self.run_command(["git", "clone",
                          "https://github.com/wlach/orangutan.git"],
                         cwd=self.workdir,
                         halt_on_failure=True)


    def patch_aosp(self):
        if self.patches != None:
            for patch in self.patches:
                projectdir = patch[0]
                url = patch[1]
                patchdir = os.path.join(self.aospdir, projectdir)
                self.info("downloading and applying AOSP patch %s to %s" % (url, patchdir))
                self.download_file(url,
                                   file_name='aosp.patch',
                                   parent_dir=self.workdir)
                self.run_command(['patch', '-p1',
                                  '-i', os.path.join(self.workdir, 'aosp.patch')],
                                 cwd=patchdir,
                                 halt_on_failure=True)

    def build_aosp(self):

        arch = None
        variant = None
        abi = None
        abi2 = ""
        if self.is_arm_target():
            arch = "arm"
            if self.is_armv7_target():
                variant = "armv7-a"
                abi = "armeabi"
                abi2 = "armeabi-v7a"
            else:
                variant = "armv5te"
                abi = "armeabi"
        else:
            arch = "x86"
            variant = "x86"
            abi = "x86"

        if abi2 != "":
            abi2 = " TARGET_CPU_ABI2=" + abi2

        env = { "BUILD_EMULATOR_OPENGL": "true",
                "BUILD_EMULATOR_OPENGL_DRIVER": "true" }

        self.run_command(["/bin/bash", "-c",
                          ". build/envsetup.sh "
                          "&& lunch sdk-eng "
                          "&& make -j " + str(self.ncores) +
                          " sdk" +
                          " TARGET_ARCH=" + arch +
                          " TARGET_ARCH_VARIANT=" + variant +
                          " TARGET_CPU_ABI=" + abi +
                          abi2 +
                          " CC=gcc-4.4 CXX=g++-4.4" +
                          " && make out/host/linux-x86/bin/mksdcard"],
                         cwd=self.aospdir,
                         halt_on_failure=True,
                         partial_env=env)

    def build_kernel(self):
        env = {}
        targ = 'goldfish_defconfig'
        if self.is_arm_target():
            env['ARCH']='arm'
            env['PATH']=(self.ndk_bin_dir() +
                         ':/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin')
            env['CROSS_COMPILE']='arm-linux-androideabi-'
            if self.is_armv7_target():
                env['SUBARCH']='armv7'
                targ='goldfish_armv7_defconfig'
            else:
                env['SUBARCH']='armv5'
        else:
            env['ARCH']='x86'
        self.run_command(["make", targ], cwd=self.goldfishdir,
                         halt_on_failure=True, partial_env=env)
        self.run_command(["make", "-j", str(self.ncores)], cwd=self.goldfishdir,
                         halt_on_failure=True, partial_env=env)


    def build_orangutan_su(self):
        self.run_command([self.ndk_bin("gcc"),
                          "--sysroot", self.ndk_sysroot(),
                          "-fPIC", "-mandroid", "-o", "su", "su.c"],
                         cwd=os.path.join(self.workdir, "orangutan"),
                         halt_on_failure=True)


    def write_registry_file(self, avddir, avdname):

        # Write the (mostly redundant) file that registers the AVD
        self.info("writing %s.ini AVD-registry file pointing to %s/avd/%s.avd"
                  % (avdname, avddir, avdname))
        reg = "avd.ini.encoding=ISO-8859-1\n"
        reg += "target=android-%s\n" % self.apilevel
        reg += "path=%s\n" % os.path.abspath(avddir)
        reg += "path.rel=avd/%s.avd\n" % avdname
        self.write_to_file(os.path.join(self.androiddir, "avd", avdname + ".ini"),
                           reg, create_parent_dir=True)

    def make_one_avd(self, avdname):

        named_avddir = os.path.join(self.avddir, avdname + ".avd")
        self.write_registry_file(named_avddir, avdname)
        self.mkdir_p(named_avddir)

        # Write the foo.avd/config.ini file that defines the AVD
        self.info("writing %s.avd/config.ini that defines the AVD" % avdname)
        ini = CONFIG_INI.copy()
        ini["image.sysdir.1"] = "platforms/android-" + self.apilevel + "/images/"
        if self.is_arm_target():
            ini["hw.cpu.arch"] = "arm"
            if self.is_armv7_target():
                ini["abi.type"] = "armeabi-v7a"
            else:
                ini["abi.type"] = "armeabi"
        else:
            ini["hw.cpu.arch"] = "x86"
            ini["abi.type"] = "x86"
            for i in ["bios.bin", "vgabios-cirrus.bin"]:
                self.copyfile(os.path.join(self.aospdir, "prebuilts/qemu-kernel/x86/pc-bios", i),
                              os.path.join(named_avddir, i))

        ini_file = open(os.path.join(named_avddir, "config.ini"), "w")
        for (k,v) in ini.items():
            ini_file.write("%s=%s\n" % (k,v))
        ini_file.close()

        # Copy the per-AVD filesystem images into place
        self.info("copying userdata images from AOSP build to AVD")
        for i in ["system.img", "ramdisk.img", "userdata.img", "userdata-qemu.img"]:
            self.copyfile(os.path.join(self.aospprodoutdir, i),
                          os.path.join(named_avddir, i))

        self.info("making 500M %s/sdcard.img" % named_avddir)
        self.run_command([os.path.join(self.aosphostbindir, "mksdcard"),
                          "500M",
                          os.path.join(named_avddir, "sdcard.img")],
                         halt_on_failure=True)

        # Copy the kernel into place
        self.info("copying kernel from goldfish build dir into %s" % self.avddir)
        kern = None
        if self.is_arm_target():
            kpath = "arch/arm/boot"
        else:
            kpath = "arch/x86/boot"
        kpath = os.path.join(self.goldfishdir, kpath)
        for i in ["zImage", "bzImage"]:
            if os.path.exists(os.path.join(kpath,i)):
                kern = os.path.join(kpath,i)
                break
        self.copyfile(os.path.join(self.goldfishdir, kern),
                      os.path.join(named_avddir, "kernel-qemu"))


    def make_base_avd(self):

        # Write the devices.xml file that defines 'mozilla-device'
        self.info("writing devices.xml that contains mozilla-device")
        self.write_to_file(os.path.join(self.androiddir, "devices.xml"),
                           MOZILLA_DEVICE_DEFINITION, create_parent_dir=True)

        self.make_one_avd("test-1")

    def emu_env(self):
        partial_env = { "ANDROID_SDK_HOME": self.workdir,
                        "ANDROID_PRODUCT_OUT": self.aospprodoutdir }
        return self.query_env(partial_env=partial_env)

    def cpu_specific_args(self, avddir):
        args = []
        if self.is_armv7_target():
            args = ["-qemu", "-cpu", "cortex-a8"]
        else:
            # point to avddir to pick up x86 BIOS
            args = ["-qemu", "-L", avddir]
        return args


    def customize_avd(self):

        self.info("starting emulator for customization run")
        args = [self.emu, "-avd", "test-1", "-no-window", "-gpu", "off", "-partition-size", "1024"]

        # unknown reason, on AWS the 64bit emulators don't seem to enjoy working;
        # for now we punt to 32bit any time we might hit a 64bit one by accident.
        if (os.path.exists(os.path.join(self.aosphostbindir, "emulator64-arm")) or
            os.path.exists(os.path.join(self.aosphostbindir, "emulator64-x86"))):
            args += ["-force-32bit"]

        # CPU-specific flags fall through to qemu, so come last
        args += self.cpu_specific_args(os.path.join(self.avddir, "test-1.avd"))

        self.info("starting emulator: " + subprocess.list2cmdline(args))
        self.info("in dir: %s" % self.aospdir)
        self.info("in env: %s" % pprint.pformat(self.emu_env()))

        p = subprocess.Popen(args,
                             cwd=self.aospdir,
                             env=self.emu_env())

        self.info("waiting for device to start")
        self.adb_e(["wait-for-device"])

        # Wait until the package manager is online as well.
        self.info("waiting for package manager to respond")
        while True:
            time.sleep(10)
            f = self.get_output_from_command([self.adb, "-e", "shell",
                                              "pm", "path", "android"],
                                             halt_on_failure=False)
            if f.startswith("package:"):
                time.sleep(10)
                break

        self.info("modifying 'su' on emulator")
        self.adb_e(["shell", "mount", "-o", "remount,rw", "/dev/block/mtdblock0", "/system"])
        self.adb_e(["push", os.path.join(self.workdir, "orangutan/su"), "/system/xbin"])
        self.adb_e(["shell", "chmod", "6755", "/system/xbin/su"])
        self.adb_e(["shell", "mount", "-o", "remount,ro", "/dev/block/mtdblock0", "/system"])

        self.info("installing packages into emulator")
        self.adb_e(["install", os.path.join(self.bindir, "Watcher.apk")])
        self.adb_e(["install", os.path.join(self.bindir, "sutAgentAndroid.apk")])

        self.info("starting watcher and SUTagent")
        self.adb_e(["shell", "am", "start", "-W", "com.mozilla.watcher/.WatcherMain"])
        self.adb_e(["shell", "am", "start", "-W", "com.mozilla.SUTAgentAndroid/.SUTAgentAndroid"])

        self.info("copying customized emulator system image from /tmp/")

        sys_image = None
        d = "/proc/%d/fd" % p.pid
        for i in os.listdir(d):
            path = os.readlink(os.path.join(d, i))
            if re.match("/tmp/android-.*/emulator-.*", path):
                if sys_image != None:
                    self.fatal("multiple plausible emulator system images, "
                               "kill emulators and try again")
                sys_image = path

        if sys_image == None:
            self.fatal("unable to find running emulator's system image")

        tmp = os.path.join(self.workdir, "system.img")
        self.copyfile(sys_image, tmp)
        self.info("terminating emulator")
        p.terminate()
        p.poll()
        self.info("moving customized system.img into %s/test-1.avd" % self.avddir)
        self.move(tmp, os.path.join(self.avddir, "test-1.avd", "system.img"))


    def adb_e(self, commands):
        self.run_command([self.adb, "-e"] + commands,
                         cwd=self.workdir,
                         halt_on_failure=True)


    def clone_customized_avd(self):
        if self.navds < 2:
            self.info("less than 2 AVDs requested, not cloning")
            return

        srcdir = os.path.join(self.avddir, "test-1.avd")
        for i in ["test-%d" % n for n in range(2,self.navds+1)]:
            self.info("making basic AVD " + i)
            self.make_one_avd(i)
            self.info("overwriting %s.avd images with customized images from test-1.avd" % i)
            dstdir = os.path.join(self.avddir, i + ".avd")
            for f in ["system.img", "userdata.img", "userdata-qemu.img"]:
                self.copyfile(os.path.join(srcdir, f),
                              os.path.join(dstdir, f))


    def bundle_avds(self):

        # clear out the cached / auto-generated .ini files in test-1.avd
        for i in ["emulator-user.ini", "hardware-qemu.ini"]:
            p = os.path.join(self.avddir, "test-1.avd", i)
            if os.path.exists(p):
                os.remove(p)

        self.info("rewriting AVD registry files to point to install target dir %s" %
                  self.config['install_android_dir'])
        for i in ["test-%d" % n for n in range(1,self.navds+1)]:
            avddir = os.path.join(self.config['install_android_dir'], "avd", i + ".avd")
            self.write_registry_file(avddir, i)

        filename = ("AVDs-%s-%s-build-%s-%s.tar.gz"
                    % (self.config['target_arch'],
                       self.tag,
                       datetime.date.today().isoformat(),
                       getpass.getuser()))
        self.run_command(["tar", "-czf", filename, "-C", self.androiddir,
                          "devices.xml", "avd"],
                         cwd=self.workdir,
                         halt_on_failure=True)


    def bundle_emulators(self):
        filename = ("emulators-%s-%s-build-%s-%s.tar.gz"
                    % (self.config['target_arch'],
                       self.tag,
                       datetime.date.today().isoformat(),
                       getpass.getuser()))
        emuarch = "x86"
        if self.is_arm_target():
            emuarch = "arm"
        self.run_command(["tar", "-czf", filename, "-C", self.aosphostdir,
                          "bin/emulator", "bin/emulator-" + emuarch, "bin/adb"],
                         cwd=self.workdir,
                         halt_on_failure=True)


    def apt_update(self):
        self.run_command(['sudo', 'apt-get', 'update'],
                         halt_on_failure=True)


    def apt_add_repo(self, repo):
        self.run_command(['sudo', 'apt-add-repository', '-y', repo],
                         halt_on_failure=True)


    def apt_get(self, pkgs):
        self.run_command(['sudo', 'apt-get', 'install', '-y'] + pkgs,
                         halt_on_failure=True)


    def is_arm_target(self):
        return self.config['target_arch'].startswith('arm')


    def is_armv7_target(self):
        return self.config['target_arch'].startswith('armv7')


    def ndk_bin_dir(self):
        if self.is_arm_target():
            return os.path.join(self.ndkdir, "toolchains/arm-linux-androideabi-4.6/prebuilt/linux-%s/bin" % self.config['host_arch'])
        else:
            return os.path.join(self.ndkdir, "toolchains/x86-4.6/prebuilt/linux-%s/bin" % self.config['host_arch'])


    def ndk_sysroot(self):
        arch = "x86"
        if self.is_arm_target():
            arch = "arm"
        apilevel = int(self.apilevel)
        while apilevel > 0:
            p = os.path.join(self.ndkdir, "platforms", "android-" + str(apilevel), "arch-" + arch)
            if os.path.exists(p):
                return p
            apilevel = int(apilevel) - 1
        self.fatal("no NDK sysroot for API level %s or less" % self.apilevel)


    def ndk_cross_prefix(self):
        if self.is_arm_target():
            return "arm-linux-androideabi-"
        else:
            return "i686-linux-android-"


    def ndk_bin(self, b):
        return os.path.join(self.ndk_bin_dir(), self.ndk_cross_prefix() + b)


    def android_apilevel(self, tag):
        pairs = [("1.6", "4"),
                 ("2.0", "5"),
                 ("2.0.1", "6"),
                 ("2.1", "7"),
                 ("2.2", "8"),
                 ("2.3", "9"),
                 ("2.3.3", "10"),
                 ("gingerbread", "10"),
                 ("3.0", "11"),
                 ("3.1", "12"),
                 ("3.2", "13"),
                 ("4.0", "14"),
                 ("4.0.3", "15"),
                 ("4.1", "16"),
                 ("4.2", "17"),
                 ("4.3", "18")]
        for (vers, api) in pairs:
            if tag == vers or tag.startswith("android-" + vers):
                self.info("android tag '%s', vers %s, uses API level %s" % (tag, vers, api))
                return api
        self.fatal("android tag '%s' doesn't map to a known API level" % tag)

    def select_android_tag(self, vers):
        tags = [
            "android-1.6_r1.1",
            "android-1.6_r1.2",
            "android-1.6_r1.3",
            "android-1.6_r1.4",
            "android-1.6_r1.5",
            "android-2.0_r1",
            "android-2.0.1_r1",
            "android-2.1_r1",
            "android-2.1_r2",
            "android-2.1_r2.1p",
            "android-2.1_r2.1s",
            "android-2.1_r2.1p2",
            "android-2.2_r1",
            "android-2.2_r1.1",
            "android-2.2_r1.2",
            "android-2.2_r1.3",
            "android-2.2.1_r1",
            "android-2.2.1_r2",
            "android-2.2.2_r1",
            "android-2.2.3_r1",
            "android-2.2.3_r2",
            "android-2.3_r1",
            "android-2.3.1_r1",
            "android-2.3.2_r1",
            "android-2.3.3_r1",
            "android-2.3.3_r1.1",
            "android-2.3.4_r0.9",
            "android-2.3.4_r1",
            "android-2.3.5_r1",
            "android-2.3.6_r0.9",
            "android-2.3.6_r1",
            "android-2.3.7_r1",
            "android-4.0.1_r1",
            "android-4.0.1_r1.1",
            "android-4.0.1_r1.2",
            "android-4.0.2_r1",
            "android-4.0.3_r1",
            "android-4.0.3_r1.1",
            "android-4.0.4_r1",
            "android-4.0.4_r1.1",
            "android-4.0.4_r1.2",
            "android-4.0.4_r2",
            "android-4.0.4_r2.1",
            "android-4.1.1_r1",
            "android-4.1.1_r1.1",
            "android-4.1.1_r2",
            "android-4.1.1_r3",
            "android-4.1.1_r4",
            "android-4.1.1_r5",
            "android-4.1.1_r6",
            "android-4.1.1_r6.1",
            "android-4.1.2_r1",
            "android-4.1.2_r2",
            "android-4.1.2_r2.1",
            "android-4.2_r1",
            "android-4.2.1_r1",
            "android-4.2.1_r1.1",
            "android-4.2.1_r1.2",
            "android-4.2.2_r1",
            "android-4.2.2_r1.1",
            "android-4.2.2_r1.1",
            "android-4.3_r0.9",
            "android-4.3_r0.9.1",
            "android-4.3_r1",
            "android-4.3_r1.1",
            "android-4.3_r2",
            "android-4.3_r2.1",
            "android-4.3_r2.2",
            "android-4.3_r2.3",
            "android-4.3_r3",
            "android-4.3.1_r1",
        ]
        codenames = {
            "jb": "4.3",
            "jellybean": "4.3",
            "ics": "4.0",
            "gb": "2.3",
            "gingerbread": "2.3",
            "froyo": "2.2",
        }

        # NB: Gingerbread is special: we don't infer to the last release branch, we infer
        # to the _development_ branch 'gingerbread' because it's where the GL emulator
        # was backported to.
        if vers == 'gingerbread':
            self.info("selecting development-branch 'gingerbread' for version '2.3'")
            return 'gingerbread'

        if vers.startswith('2.3'):
            self.info("android version 2.3.x specified, when dev branch 'gingerbread' is most likely")
            self.info("required to get a working fennec build; pass --android-tag if you are certain")
            self.info("you want a non-'gingerbread' 2.3.x version")
            self.fatal("passed 2.3.x for --android-version, failing conservatively")

        if vers in codenames:
            self.info("using android version '%s' for requested codename '%s'" % (codenames[vers], vers))
            vers = codenames[vers]

        for tag in reversed(tags):
            if tag.startswith(vers) or tag.startswith("android-" + vers):
                self.info("selecting tag '%s' for version '%s'" % (tag, vers))
                return tag
        self.fatal("requested android version '%s' doesn't match any tag" % vers)

    def select_patches(self, tag):
        if tag == 'gingerbread':
            # FIXME: perhaps put this patch someplace more stable than bugzilla?
            return [('development',
                     'https://bug910092.bugzilla.mozilla.org/attachment.cgi?id=8361456'),
                    ('external/sqlite',
                     'https://bug910092.bugzilla.mozilla.org/attachment.cgi?id=8364687')]
        return None

    def _post_fatal(self, message=None, exit_code=None):
        for i in ['emulator', 'emulator-arm', 'emulator-x86',
                  'emulator64', 'emulator64-arm', 'emulator64-x86']:
            self._kill_processes(i)


# main
if __name__ == '__main__':
    myScript = EmulatorBuild()
    myScript.run_and_exit()
